# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from enum import Enum
from functools import singledispatchmethod
from numbers import Number
from pydoc import locate
from types import NoneType

import numpy as np

import qat.ir.instructions as pyd_instructions
import qat.ir.measure as pyd_measure
import qat.ir.waveforms as pyd_waveforms
from qat.backend.passes.purr.analysis import PulseChannelTimeline
from qat.core.pass_base import TransformPass
from qat.core.result_base import ResultInfoMixin, ResultManager
from qat.ir.instruction_builder import PydQuantumInstructionBuilder
from qat.ir.lowered import PartitionedIR
from qat.middleend.passes.analysis import ActivePulseChannelResults
from qat.middleend.passes.purr.analysis import ActiveChannelResults
from qat.middleend.passes.purr.transform import LoopCount
from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, PulseChannelView, PulseShapeType, Qubit
from qat.purr.compiler.hardware_models import HardwareModel
from qat.purr.compiler.instructions import (
    BinaryOperator,
    CustomPulse,
    Instruction,
    InstructionBlock,
    Jump,
    PostProcessing,
    Pulse,
    QuantumInstruction,
    Repeat,
    Reset,
    ResultsProcessing,
    Synchronize,
    Variable,
)


class ConvertToPydanticIR(TransformPass):
    def __init__(
        self,
        legacy_model: HardwareModel,
        pyd_model: PhysicalHardwareModel,
        *args,
        **kwargs,
    ):
        self.legacy_model = legacy_model
        self.pyd_model = pyd_model
        self._create_pulse_channel_map()
        self._additional_data: dict = {}

    def _create_pulse_channel_map(self):
        """Create a mapping of legacy pulse channels to Pydantic pulse channels."""
        self._pulse_channel_map = {}
        self._pulse_channel_to_qubit_index_map = {}
        for index, qubit in self.pyd_model.qubits.items():
            q_key = f"Q{index}"
            self._pulse_channel_map[q_key + ".drive"] = qubit.drive_pulse_channel.uuid
            self._pulse_channel_to_qubit_index_map[q_key + ".drive"] = index
            self._pulse_channel_map[q_key + ".freq_shift"] = (
                qubit.freq_shift_pulse_channel.uuid
            )
            self._pulse_channel_map[q_key + ".second_state"] = (
                qubit.second_state_pulse_channel.uuid
            )
            for other_index, channel in qubit.cross_resonance_pulse_channels.items():
                self._pulse_channel_map[f"{q_key}.Q{other_index}.cross_resonance"] = (
                    channel.uuid
                )
            for (
                other_index,
                channel,
            ) in qubit.cross_resonance_cancellation_pulse_channels.items():
                self._pulse_channel_map[
                    f"{q_key}.Q{other_index}.cross_resonance_cancellation"
                ] = channel.uuid
            resonator = qubit.resonator
            r_key = f"R{index}"
            self._pulse_channel_map[r_key + ".measure"] = (
                resonator.measure_pulse_channel.uuid
            )
            self._pulse_channel_map[r_key + ".acquire"] = (
                resonator.acquire_pulse_channel.uuid
            )

    def run(
        self,
        ir: InstructionBuilder | PartitionedIR,
        res_mgr: ResultManager,
        *args,
        **kwargs,
    ):
        self._additional_data = {}
        if isinstance(ir, PartitionedIR):
            new_ir = self._convert_partitioned_ir(ir)
        elif isinstance(ir, InstructionBuilder):
            new_ir = self._convert_instruction_builder(ir)
        else:
            raise TypeError(f"Unsupported IR type: {type(ir)}")
        for key, value in self._additional_data.items():
            setattr(new_ir, key, value)
        self._convert_result_manager(res_mgr)
        return new_ir

    def _convert_partitioned_ir(self, ir: PartitionedIR) -> PartitionedIR:
        new_ir = PartitionedIR()
        for attr, value in vars(ir).items():
            setattr(new_ir, attr, self._convert_element(value))
        return new_ir

    def _convert_instruction_builder(
        self, ir: InstructionBuilder
    ) -> PydQuantumInstructionBuilder:
        # TODO: Implement full compatibility of InstructionBuilders
        # Variables: COMPILER-589
        new_ir = PydQuantumInstructionBuilder(
            hardware_model=self.pyd_model,
        )
        for attr, value in vars(ir).items():
            if attr == "_instructions":
                attr = "instructions"
            setattr(new_ir, attr, self._convert_element(value))
        return new_ir

    def _convert_result_manager(self, res_mgr: ResultManager):
        """Convert the ResultManager to a Pydantic-compatible format."""
        if res_mgr is None:
            return
        if not isinstance(res_mgr, ResultManager):
            raise TypeError(f"Unsupported ResultManager type: {type(res_mgr)}")
        new_res_mgr = ResultManager()
        for result in res_mgr._results:
            new_res_mgr.add(self._convert_element(result._result))
        res_mgr._results.clear()
        res_mgr.update(new_res_mgr)
        return

    @singledispatchmethod
    def _convert_element(self, value, *args, **kwargs):
        """Default method for converting elements. This will raise an error if no specific
        conversion is defined for the type of `value`.
        """
        raise TypeError(f"Unsupported type for conversion: {type(value)}")

    @_convert_element.register(ResultInfoMixin)
    def _(
        self,
        value: ResultInfoMixin,
        *args,
        **kwargs,
    ):
        """Convert a ResultInfoMixin instance."""
        data = dict()
        kwargs["as_index"] = False
        for attr, val in vars(value).items():
            data[attr] = self._convert_element(val, *args, **kwargs)

        class_name = value.__class__.__name__
        # Check non-legacy path for class name first, if not found, use legacy path
        class_path = value.__class__.__module__.replace(".purr", "") + "." + class_name
        cls_ = locate(class_path) or value.__class__
        return cls_(**data)

    @_convert_element.register(ActiveChannelResults)
    def _(
        self,
        value: ActiveChannelResults,
        *args,
        **kwargs,
    ):
        """Convert an ActiveChannelResults instance to ActivePulseChannelResults."""
        result = ActivePulseChannelResults()
        for pulse_channel, device in value.target_map.items():
            pulse_channel_id = self._convert_element(pulse_channel)
            device = self._convert_element(device)
            result.pulse_channel_to_qubit_map[pulse_channel_id] = device
            result.qubit_to_pulse_channel_map[device].append(pulse_channel_id)
        return result

    @_convert_element.register(PulseChannelTimeline)
    def _(
        self,
        value: PulseChannelTimeline,
        *args,
        **kwargs,
    ):
        """Convert a PulseChannelTimeline instance."""
        for attr, val in vars(value).items():
            setattr(value, attr, self._convert_element(val))
        return value

    @_convert_element.register(type)
    def _(
        self,
        value: type,
        *args,
        **kwargs,
    ):
        """Convert a type to its Pydantic equivalent."""
        pyd_class = self._get_pyd_class(value.__name__, allow_none=True)
        if pyd_class is None:
            return value
        return pyd_class

    @_convert_element.register(Number)
    @_convert_element.register(str)
    @_convert_element.register(NoneType)
    @_convert_element.register(np.ndarray)
    def _(
        self,
        value: Number | str | None,
        *args,
        **kwargs,
    ):
        return value

    @_convert_element.register(LoopCount)
    def _(
        self,
        value: LoopCount,
        *args,
        **kwargs,
    ):
        """Convert a LoopCount instance."""
        return pyd_instructions.LoopCount(value)

    @_convert_element.register(list)
    @_convert_element.register(tuple)
    @_convert_element.register(set)
    def _(
        self,
        value: list | tuple | set,
        *args,
        **kwargs,
    ):
        if all(isinstance(v, (Number, str)) for v in value):
            return value
        else:
            temp_list = []
            for item in value:
                new_item = self._convert_element(item)
                if isinstance(new_item, list) and not isinstance(item, list):
                    # If the item is a list and the original value is not a list,
                    # we need to flatten it.
                    temp_list.extend(new_item)
                else:
                    temp_list.append(new_item)
            return type(value)(temp_list)

    @_convert_element.register(dict)
    def _(
        self,
        value: dict,
        *args,
        **kwargs,
    ):
        """Convert a dictionary by iterating over its items."""
        new_dict = {}
        for key, val in value.items():
            new_key = self._convert_element(key, *args, **kwargs)
            if isinstance(val, (int, float, str)):
                new_dict[new_key] = val
            else:
                new_dict[new_key] = self._convert_element(val, *args, **kwargs)
        return new_dict

    @_convert_element.register(HardwareModel)
    def _(
        self,
        value: HardwareModel,
        *args,
        **kwargs,
    ):
        return self.pyd_model

    @_convert_element.register(Variable)
    def _(
        self,
        value: Variable,
        *args,
        **kwargs,
    ):
        """Convert a Variable instance."""
        if value.var_type is LoopCount and isinstance(value.value, int):
            value.value = LoopCount(value.value)
        new_var = pyd_instructions.Variable(
            name=value.name,
            var_type=self._convert_element(value.var_type, *args, **kwargs),
            value=self._convert_element(value.value, *args, **kwargs),
        )
        return new_var

    @_convert_element.register(PulseChannelView)
    @_convert_element.register(PulseChannel)
    def _(
        self,
        value: PulseChannelView | PulseChannel,
        *args,
        **kwargs,
    ):
        """Convert a PulseChannelView instance."""
        return self._pulse_channel_map[value.id]

    @_convert_element.register(Qubit)
    def _(
        self,
        value: Qubit,
        *args,
        as_index: bool = True,
        **kwargs,
    ):
        """Convert a Qubit instance."""
        if as_index:
            return value.index
        return self.pyd_model.qubit_with_index(value.index)

    @_convert_element.register(Instruction)
    @_convert_element.register(BinaryOperator)
    def _(
        self,
        value: Instruction | BinaryOperator,
        *args,
        **kwargs,
    ):
        """Convert an Instruction instance."""
        data = dict()
        for name, var in vars(value).items():
            if name == "quantum_targets":
                data["targets"] = frozenset(self._convert_element(var, *args, **kwargs))
            elif name == "time":
                data["duration"] = self._convert_element(var, *args, **kwargs)
            else:
                data[name] = self._convert_element(var, *args, **kwargs)
        pyd_class = self._get_pyd_class(value)
        return pyd_class(**data)

    @_convert_element.register(Synchronize)
    def _(self, value: Synchronize, *args, **kwargs):
        """Convert a Synchronize instance."""
        data = dict()
        for name, var in vars(value).items():
            if name == "quantum_targets":
                data["targets"] = frozenset(self._convert_element(var, *args, **kwargs))
                if len(data["targets"]) <= 1:
                    return []
            else:
                data[name] = self._convert_element(var, *args, **kwargs)
        pyd_class = self._get_pyd_class(value)
        return pyd_class(**data)

    @_convert_element.register(QuantumInstruction)
    def _(self, value: QuantumInstruction, *args, **kwargs):
        """Convert a QuantumInstruction instance."""
        data = dict()
        for name, var in vars(value).items():
            if name == "quantum_targets":
                targets = self._convert_element(var, *args, **kwargs)
            elif name == "time":
                data["duration"] = self._convert_element(var, *args, **kwargs)
            else:
                data[name] = self._convert_element(var, *args, **kwargs)
        pyd_class = self._get_pyd_class(value)
        return [pyd_class(target=target, **data) for target in targets]

    @_convert_element.register(Reset)
    def _(self, value: Reset, *args, **kwargs):
        """Convert a Reset instance."""
        qubit_targets = set(
            [
                self._pulse_channel_to_qubit_index_map[target.id]
                for target in value.quantum_targets
            ]
        )
        duration = self._convert_element(value.duration, *args, **kwargs)
        return [
            pyd_instructions.Reset(
                qubit_target=target,
                duration=duration,
            )
            for target in qubit_targets
        ]

    @_convert_element.register(InstructionBlock)
    def _(
        self,
        value: InstructionBlock,
        *args,
        **kwargs,
    ):
        """Convert an InstructionBlock instance."""
        pyd_class = self._get_pyd_class(value)
        extra_data = {}
        if "qubit_targets" in pyd_class.model_fields:
            extra_data["qubit_targets"] = frozenset(
                self._convert_element(value.quantum_targets, *args, **kwargs)
            )
        new_block = pyd_class(**extra_data)
        for instruction in value.instructions:
            new_block.add(*self._convert_element(instruction, *args, **kwargs))
        return new_block

    @_convert_element.register(Repeat)
    def _(
        self,
        value: Repeat,
        *args,
        **kwargs,
    ):
        """Convert a Repeat instance."""
        data = dict()
        for name, var in vars(value).items():
            if name == "repeat_count":
                repeat = pyd_instructions.Repeat(
                    repeat_count=self._convert_element(var, *args, **kwargs)
                )
            else:
                data[name] = self._convert_element(var, *args, **kwargs)
        self._additional_data.update(data)
        return repeat

    @_convert_element.register(ResultsProcessing)
    @_convert_element.register(PostProcessing)
    def _(
        self,
        value: ResultsProcessing | PostProcessing,
        *args,
        **kwargs,
    ):
        """Convert an Assign instance."""
        return self._get_pyd_class(value)._from_legacy(value)

    @_convert_element.register(Jump)
    def _(
        self,
        value: Jump,
        *args,
        **kwargs,
    ):
        """Convert a Jump instance."""
        return pyd_instructions.Jump(
            label=self._convert_element(value.target, *args, **kwargs),
            condition=self._convert_element(value.condition, *args, **kwargs),
        )

    @_convert_element.register(Enum)
    def _(
        self,
        value: Enum,
        *args,
        **kwargs,
    ):
        """Convert an Enum instance."""
        pyd_class = self._get_pyd_class(value)
        return pyd_class(value.value)

    @_convert_element.register(PulseShapeType)
    def _(
        self,
        value: PulseShapeType,
        *args,
        **kwargs,
    ):
        """Convert a PulseShapeType instance."""
        if value.name == "GAUSSIAN_DRAG":
            class_name = "DragGaussianWaveform"
        else:
            class_name = f"{value.name.title().replace('_', '')}Waveform"
        pyd_class = getattr(pyd_waveforms, class_name, None)
        if pyd_class is None:
            raise TypeError(f"Unsupported PulseShapeType: {value}")
        return pyd_class

    @_convert_element.register(CustomPulse)
    def _(
        self,
        value: CustomPulse,
        *args,
        **kwargs,
    ):
        """Convert a CustomPulse instance."""
        return pyd_waveforms.Pulse(
            waveform=pyd_waveforms.SampledWaveform(samples=value.samples),
            targets=frozenset(
                self._convert_element(value.quantum_targets, *args, **kwargs)
            ),
            ignore_channel_scale=value.ignore_channel_scale,
            duration=value.duration,
        )

    @_convert_element.register(Pulse)
    def _(
        self,
        value: Pulse,
        *args,
        **kwargs,
    ):
        """Convert a Pulse instance."""
        waveform_data = dict()
        pulse_data = dict()
        for name, var in vars(value).items():
            if name == "quantum_targets":
                pulse_data["targets"] = frozenset(
                    self._convert_element(var, *args, **kwargs)
                )
            elif name in ("ignore_channel_scale", "duration"):
                pulse_data[name] = var
            else:
                waveform_data[name] = self._convert_element(var, *args, **kwargs)
        waveform_class = waveform_data.pop("shape")
        return pyd_waveforms.Pulse(
            waveform=waveform_class(**waveform_data),
            **pulse_data,
        )

    @staticmethod
    def _get_pyd_class(val: object | str, allow_none: bool = False) -> type | None:
        """Get the Pydantic class corresponding to the legacy instruction name."""
        name = val.__class__.__name__ if not isinstance(val, str) else val
        pyd_class = getattr(pyd_instructions, name, None)
        if pyd_class is None:
            pyd_class = getattr(pyd_measure, name, None)
        if pyd_class is None and not allow_none:
            raise TypeError(f"Unsupported type: {name}")
        return pyd_class



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import abc
import itertools as it
from typing import Any, Iterable

from compiler_config.config import InlineResultsProcessing, Languages
from pydantic import Field

from qat.frontend.register import BitRegister, CregIndexValue, QubitRegister, Registers
from qat.ir.instruction_builder import QuantumInstructionBuilder
from qat.ir.instructions import Variable
from qat.model.device import Qubit
from qat.purr.utils.logger import get_default_logger
from qat.utils.pydantic import NoExtraFieldsModel, ValidatedDict, ValidatedList

log = get_default_logger()


class QasmContext(NoExtraFieldsModel):
    """
    Container object for all data relating to the scope/pass of QASM currently under
    analysis.
    """

    registers: Registers = Registers()
    variables: ValidatedDict[str, Variable] = Field(
        default_factory=lambda: ValidatedDict[str, Variable]()
    )
    gates: dict[str, Any] = Field(default_factory=dict)


class ParseResults(NoExtraFieldsModel):
    """
    Results object for attempted parse. When coerced to a boolean matches
    against if parse was successful.
    """

    success: bool
    errors: str = ""

    def __bool__(self):
        return self.success


class AbstractParser(abc.ABC):
    def __init__(self):
        self.results_format = InlineResultsProcessing.Program
        self._cached_parses: dict[int, Any] = dict()

    def can_parse(self, qasm: str) -> ParseResults:
        try:
            self._fetch_or_parse(qasm)
            return ParseResults(success=True)
        except Exception as ex:
            return ParseResults(success=False, errors=str(ex))

    def parser_language(self) -> Languages:
        return Languages.Empty

    @abc.abstractmethod
    def _fetch_or_parse(self): ...

    @abc.abstractmethod
    def walk_node(self, node, context: QasmContext, builder, **kwargs):
        pass

    def _add_delay(
        self, delay: float, qubits: list[Qubit], builder: QuantumInstructionBuilder
    ):
        registers = self._expand_to_match_registers(qubits, flatten_results=True)
        for qubit in registers:
            builder.delay(qubit, delay)

    def _add_qreg(
        self,
        reg_name: str,
        reg_length: int,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
    ):
        if reg_name in context.registers.quantum:
            return
        index_range = self._get_qreg_index_range(reg_length, context, builder)
        qubits = []
        for idx in index_range:
            qubit = builder.hw.qubit_with_index(idx)
            qubits.append(qubit)
        context.registers.quantum[reg_name] = QubitRegister(qubits=qubits)

    def _add_creg(self, reg_name: str, reg_length: int, context: QasmContext):
        if reg_name in context.registers.classic:
            return
        context.registers.classic[reg_name] = BitRegister(
            bits=[
                CregIndexValue(register_name=reg_name, index=idx, value=0)
                for idx in range(reg_length)
            ]
        )

    def _add_measure(
        self,
        qubits: list[Qubit],
        clbits: list[CregIndexValue],
        builder: QuantumInstructionBuilder,
    ):
        registers = self._expand_to_match_registers(qubits, clbits, flatten_results=True)

        for qubit, creg in registers:
            qubit_idx = builder._qubit_index_by_uuid[qubit.uuid]

            res_id = f"{creg}_{qubit_idx}"
            builder.measure_single_shot_z(qubit, output_variable=res_id)
            creg.value = res_id
            builder.results_processing(res_id, self.results_format)

    def _add_unitary(
        self,
        theta: float,
        phi: float,
        lamb: float,
        qubit_or_register: list[Qubit | QubitRegister],
        builder: QuantumInstructionBuilder,
    ):
        """Unitary in QASM terms is just ``U(...)``."""
        qubits = self._expand_to_match_registers(qubit_or_register, flatten_results=True)
        for qubit in qubits:
            builder.Z(qubit, lamb).Y(qubit, theta).Z(qubit, phi)

    def _add_cnot(
        self,
        control_qbs: list[Qubit],
        target_qbs: list[Qubit],
        builder: QuantumInstructionBuilder,
    ):
        for cqb, tqb in self._expand_to_match_registers(
            control_qbs, target_qbs, flatten_results=True
        ):
            builder.cnot(cqb, tqb)

    def _add_reset(self, qubits: list[Qubit], builder: QuantumInstructionBuilder):
        qubits = self._expand_to_match_registers(qubits, flatten_results=True)
        builder.reset(qubits)

    def _add_if(self, left, right, if_body, context, builder):
        raise NotImplementedError("Conditional logic is currently not implemented.")

    def _add_ecr(self, qubits: list[Qubit], builder: QuantumInstructionBuilder):
        if len(qubits) != 2:
            raise ValueError(
                "Definition of ECR intrinsic is wrong. Can only take 2 qubits as arguments."
            )
        builder.ECR(qubits[0], qubits[1])

    def _get_qreg_index_range(
        self, reg_length: int, context: QasmContext, builder: QuantumInstructionBuilder
    ):
        next_free = 0
        available_indices = list(builder.hw.qubits.keys())
        max_used = max(
            [-1]
            + [
                builder._qubit_index_by_uuid[qubit.uuid]
                for qubit_reg in context.registers.quantum.values()
                for qubit in qubit_reg.qubits
            ]
        )

        if max_used > -1:
            next_free = available_indices.index(max_used) + 1
        index_range = available_indices[next_free : next_free + reg_length]
        if len(index_range) < reg_length:
            raise ValueError("Attempted to allocate more qubits than available.")

        return index_range

    def _is_register_target(self, values: list[Any]):
        """
        Does it look like the passed-in qubit or parameter list contains a register.
        """
        return isinstance(values, Iterable) and any(
            isinstance(val, (QubitRegister, BitRegister)) for val in values
        )

    def _curate_register_input(self, input: list | Qubit) -> list:
        """
        If the input is a list of (qu)bit registers, curate the input such that
        we are left with a list of (qu)bits. Do nothing if a list of (qu)bits is passed.
        """
        if isinstance(input, list | ValidatedList):
            if len(input) == 1 and isinstance(
                register := input[0], QubitRegister | BitRegister
            ):
                return register.contents
            else:
                return list(input)
        elif isinstance(input, Qubit):
            return [input]
        else:
            raise TypeError("Please provide a list as input.")

    def _expand_to_match_registers(self, *args, tuple_return=True, flatten_results=False):
        """
        Expands and zips registers/non-registers together so they can be processed.

        QASM treats any registers of bits or qubits as calling the gate * register size
        times with each individual register value. This is a helper method for
        centralizing this expansion/mapping.

        :Example:

        .. code-block:: python

            [[q1, q2, q3], [q4, q5, q6],[c1]]

        With the first 2 lists being expanded qubit registers, should resolve into:

        .. code-block:: python

            [(q1, q4, c1), (q2, q5, c1), (q3, q6, c1)]
        """
        curated_args = []
        for arg in args:
            curated_args.append(self._curate_register_input(arg))

        if len(curated_args) > 1:
            # Each list needs to be of equal length, repeat element if necessary.
            curated_args = [
                arg if len(arg) > 1 else arg * len(max(curated_args, key=len))
                for arg in curated_args
            ]
            return list(it.zip_longest(*curated_args))
        else:
            return curated_args[0]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from compiler_config.config import InlineResultsProcessing
from pydantic import BaseModel, PositiveInt, field_validator
from pydantic_core import from_json

from qat.ir.instructions import Assign
from qat.ir.measure import PostProcessing
from qat.purr.compiler.instructions import AcquireMode
from qat.utils.pydantic import RehydratableModel


class AcquireData(BaseModel):
    """Holds the information required at runtime surrounding an acquisition.

    :param mode: The type of acquisition performed on the hardware.
    :param shape: The shape the data is expected to be returned in. Reflects the type of
        acquisition (e.g. scope or integrator mode), as well as the iteration structure
        (e.g. shots, loops). A shape of (1000,) would indicate 1000 shots of a single
        integrator acquisition.
    :param physical_channel: The physical channel the acquisition is performed on, needed
        for post-processing reasons (e.g. error mitigation).
    :param post_processing: A list of post-processing steps to be applied to the execution
        data.
    :param results_processing: Contains information surrounding how the results should be
        formatted after post-processing.
    """

    mode: AcquireMode
    shape: tuple[int, ...]
    physical_channel: str
    post_processing: list[PostProcessing] = []
    results_processing: InlineResultsProcessing | None = None


class AbstractProgram(RehydratableModel, ABC):
    """Abstract base class for programs to be executed by an engine.

    Child classes must implement the :attr:`acquire_shapes` property, which returns a
    mapping of acquisition output variables to the shape of the acquisition for this
    program. This specification enforces programs to be explicit about what acquisition
    outputs they will produce.
    """

    @property
    @abstractmethod
    def acquire_shapes(self) -> dict[str, tuple[int, ...]]: ...


Program = TypeVar("Program", bound=AbstractProgram)


class Executable(BaseModel, Generic[Program]):
    """:class:`Executable`s are packages of instructions that will be executed by the
    runtime. They contain a :class:`Program`, or a number of :class:`Program`s to be
    executed by the engine. They also contain the acquire information, such as
    post-processing and result assignments.

    :param programs: The program(s) to be executed against the engine.
    :param assigns: Assigns results to given variables.
    :param returns: Which acqusitions/variables should be returned.
    :param calibration_id: The (unique) cabliration ID of the underlying hardware model.
    :param shots: The total number of shots performed in the :class:`Executable`.
    """

    programs: list[Program] = []
    acquires: dict[str, AcquireData] = dict()
    assigns: list[Assign] = []
    returns: set[str] = set()
    calibration_id: str = ""
    shots: PositiveInt | None = None

    def serialize(self, indent: int = 4) -> str:
        """Serializes the executable as a JSON blob."""
        return self.model_dump_json(indent=indent, exclude_none=True)

    @classmethod
    def deserialize(cls, blob: str):
        """Instantiates a executable from a JSON blob."""
        return cls(**from_json(blob))

    @field_validator("programs", mode="before")
    @classmethod
    def _rehydrate_programs(cls, data) -> list[Program]:
        """This validator determines the correct type of program to rehydrate based on the
        input data structure."""
        if isinstance(data, list):
            return [
                AbstractProgram._rehydrate_object(d) if isinstance(d, dict) else d
                for d in data
            ]
        else:
            return cls._rehydrate_programs([data])

    @field_validator("programs", mode="after")
    @classmethod
    def _validate_programs_are_the_same_type(cls, programs):
        """Ensures that if multiple programs are provided, they are all of the same type."""
        if isinstance(programs, list):
            types = set(type(p) for p in programs)
            if len(types) > 1:
                raise ValueError(
                    f"All programs in the executable must be of the same type. Saw {types}."
                )
        return programs



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.loaders.purr.qiskit import QiskitModelLoader
from qat.pipelines.updateable import PipelineConfig

from .compile import LegacyQiskitCompilePipeline
from .execute import LegacyQiskitExecutePipeline
from .full import LegacyQiskitPipeline


def _create_pipeline_instance(num_qubits: int) -> LegacyQiskitPipeline:
    return LegacyQiskitPipeline(
        config=PipelineConfig(name=f"legacy_qiskit{num_qubits}"),
        loader=QiskitModelLoader(qubit_count=num_qubits),
    ).pipeline


legacy_qiskit8 = _create_pipeline_instance(8)
legacy_qiskit16 = _create_pipeline_instance(16)
legacy_qiskit32 = _create_pipeline_instance(32)

__all__ = [
    "LegacyQiskitPipeline",
    "LegacyQiskitCompilePipeline",
    "LegacyQiskitExecutePipeline",
    "PipelineConfig",
    "legacy_qiskit8",
    "legacy_qiskit16",
    "legacy_qiskit32",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import hashlib
import itertools as it
import random
from collections import defaultdict
from pathlib import Path

import networkx as nx
import numpy as np

from qat.model.builder import PhysicalHardwareModelBuilder
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.compiler.devices import (
    ChannelType,
    PhysicalBaseband,
    PhysicalChannel,
    PulseShapeType,
    Qubit,
    QubitCoupling,
    Resonator,
)
from qat.purr.compiler.hardware_models import (
    ErrorMitigation,
    QuantumHardwareModel,
    ReadoutMitigation,
)
from qat.utils.pydantic import CalibratableUnitInterval2x2Array


def random_connectivity(n, max_degree=3, seed=42):
    """
    Generates a random undirected graph but enforcing that the resulting graph is connected.
    """
    seeded_random = seed if isinstance(seed, random.Random) else random.Random(seed)
    edges = list(it.combinations(range(n), 2))
    seeded_random.shuffle(edges)
    G = nx.Graph()
    G.add_nodes_from(range(n))
    for node_edges in edges:
        if (
            len(G.edges(node_edges[0])) < max_degree
            and len(G.edges(node_edges[1])) < max_degree
        ):
            G.add_edge(*node_edges)

    return {node: set(neighbors) for node, neighbors in G.adjacency()}


def random_directed_connectivity(n, max_degree=3, seed=42):
    """
    Generates a random directed graph but enforcing that the resulting graph is connected.
    """
    edges = list(it.combinations(range(n), 2))
    random.Random(seed).shuffle(edges)
    G = nx.DiGraph()
    G.add_nodes_from(range(n))
    for node_edges in edges:
        if (
            len(G.edges(node_edges[0])) < max_degree
            and len(G.edges(node_edges[1])) < max_degree
        ):
            G.add_edge(*node_edges)

    return {node: set(neighbors) for node, neighbors in G.adjacency()}


def random_quality_map(connectivity, seed=42, min_quality=0.0, max_quality=1.0):
    seeded_random = seed if isinstance(seed, random.Random) else random.Random(seed)
    coupling_map = {}
    for q1_index, connected_qubits in connectivity.items():
        for q2_index in connected_qubits:
            coupling_map[(q1_index, q2_index)] = seeded_random.uniform(
                min_quality, max_quality
            )
    return coupling_map


def random_error_mitigation(physicaal_indices, seed: int | None = None) -> ErrorMitigation:
    seeded_random = random.Random(seed)
    linear: dict = {}
    for q_id in physicaal_indices:
        p00 = seeded_random.uniform(0.8, 1.0)
        p11 = seeded_random.uniform(0.8, 1.0)
        linear[str(q_id)] = {
            "0|0": p00,
            "1|0": 1 - p00,
            "1|1": p11,
            "0|1": 1 - p11,
        }

    return ErrorMitigation(readout_mitigation=ReadoutMitigation(linear=linear))


def pick_subconnectivity(connectivity, n, seed=42):
    seeded_random = seed if isinstance(seed, random.Random) else random.Random(seed)
    sub_qubits = seeded_random.sample(list(connectivity.keys()), n)
    sub_connectivity = defaultdict(set)
    for qubit in sub_qubits:
        for connected_qubit in connectivity[qubit]:
            if qubit not in sub_connectivity[connected_qubit]:
                # TODO: Discussed that this should possibly be something like:
                # if connected_qubit in sub_qubits:
                # Evaluate if it should be changed
                # COMPILER-822
                sub_connectivity[qubit].add(connected_qubit)

    return sub_connectivity


def ensure_connected_connectivity(connectivity: dict, qubit_indices: list | set) -> dict:
    """Ensures that all `qubit_indices` are connected together.

    Looks at only the selected qubit indices and ensures that they are all part of a
    single connected graph. This does not mean an all-to-all connection.
    E.g. While the connectivity dict below has all qubits linked, in the subset of indices
    `{1, 2, 3}` qubit `0` is not linked with the other two.

    .. code:: python

        connectivity = {
            0: {1, 2},
            1: {0},
            2: {0, 3},
            3: {2},
        }
        qubit_indices = {1, 2, 3}
        new_connectivity = ensure_connected_connectivity(connectivity, qubit_indices)
        # new_connectivity == {
        #     0: {1, 2},
        #     1: {0, 2},
        #     2: {0, 1, 3},
        #     3: {2}
        # }

    :param connectivity: Base connectivity dictionary.
    :param qubit_indices: Selected qubits to ensure are a connected graph.
    """
    new_connectivity = {k: set(v) for k, v in connectivity.items()}
    G = nx.Graph()
    G.add_nodes_from(qubit_indices)
    for q1_index in qubit_indices:
        for q2_index in filter(
            lambda x: x in qubit_indices, new_connectivity.get(q1_index, [])
        ):
            G.add_edge(q1_index, q2_index)
    if not nx.is_connected(G):
        connected_generators = nx.connected_components(G)
        main_component = next(connected_generators)
        tail = list(main_component)[-1]
        for component in connected_generators:
            head = list(component)[0]
            new_connectivity[tail].add(head)
            new_connectivity[head].add(tail)
            G.add_edge(tail, head)
            G.add_edge(head, tail)
            tail = list(component)[-1]
        if not nx.is_connected(G):
            raise ValueError("The provided connectivity is not connected.")
    return new_connectivity


def generate_connectivity_data(n_qubits, n_logical_qubits, seed=42):
    physical_connectivity = random_connectivity(n=n_qubits, seed=seed)
    logical_connectivity = pick_subconnectivity(
        physical_connectivity, n=n_logical_qubits, seed=seed
    )
    logical_connectivity_quality = random_quality_map(
        connectivity=logical_connectivity, seed=seed
    )
    return (physical_connectivity, logical_connectivity, logical_connectivity_quality)


def generate_hw_model(n_qubits, seed=42):
    physical_connectivity, _, _ = generate_connectivity_data(n_qubits, n_qubits, seed=seed)

    builder = PhysicalHardwareModelBuilder(physical_connectivity=physical_connectivity)
    return builder.model


def apply_setup_to_echo_hardware(
    qubit_count: int, connectivity, qubit_indices: set = None
) -> QuantumHardwareModel:
    qubit_devices = []
    resonator_devices = []
    channel_index = 1
    hw = QuantumHardwareModel()
    if qubit_indices is not None:
        if (no_qubits_indices := len(qubit_indices)) > qubit_count:
            qubit_indices = it.islice(qubit_indices, qubit_count)
        elif no_qubits_indices < qubit_count:
            raise ValueError(
                f"Not enough qubit indices provided: len({qubit_indices}) = {no_qubits_indices} < {qubit_count}."
            )
    else:
        qubit_indices = range(qubit_count)
    for primary_index in qubit_indices:
        bb1 = PhysicalBaseband(f"LO{channel_index}", 5.5e9)
        bb2 = PhysicalBaseband(f"LO{channel_index + 1}", 8.5e9)
        hw.add_physical_baseband(bb1, bb2)

        ch1 = PhysicalChannel(f"CH{channel_index}", 1.0e-9, bb1, 1)
        ch2 = PhysicalChannel(
            f"CH{channel_index + 1}", 1.0e-9, bb2, 1, acquire_allowed=True
        )
        hw.add_physical_channel(ch1, ch2)

        resonator = Resonator(f"R{primary_index}", ch2)
        resonator.create_pulse_channel(ChannelType.measure, frequency=8.5e9)
        resonator.create_pulse_channel(ChannelType.acquire, frequency=8.5e9)

        qubit = Qubit(primary_index, resonator, ch1)
        qubit.create_pulse_channel(ChannelType.drive, frequency=5.5e9)
        qubit.pulse_hw_x_pi_2 = {
            "shape": PulseShapeType.GAUSSIAN,
            "width": random.Random().uniform(50e-9, 1000e-9),
            "rise": random.Random().uniform(1 / 6, 1 / 2),
            "amp": random.Random().uniform(1e6, 5e6),
            "drag": random.Random().uniform(0, 0.5),
            "phase": random.Random().uniform(0, 1),
        }

        qubit_devices.append(qubit)
        resonator_devices.append(resonator)
        channel_index = channel_index + 2

    qubits_by_index = {qb.index: qb for qb in qubit_devices}
    if isinstance(connectivity, list):
        connectivity = dict.fromkeys(connectivity, None)
    for connection, quality in connectivity.items():
        left_index, right_index = connection
        qubit_left = qubits_by_index.get(left_index, None)
        qubit_right = qubits_by_index.get(right_index, None)

        if not all([qubit_left, qubit_right]):
            continue

        qubit_left.create_pulse_channel(
            auxiliary_devices=[qubit_right],
            channel_type=ChannelType.cross_resonance,
            frequency=5.5e9,
            scale=np.complex128(50 + 50j),
        )
        qubit_left.create_pulse_channel(
            auxiliary_devices=[qubit_right],
            channel_type=ChannelType.cross_resonance_cancellation,
            frequency=5.5e9,
            scale=np.complex128(5 + 5j),
        )
        qubit_left.add_coupled_qubit(qubit_right)
        qubit_left.pulse_hw_zx_pi_4[qubit_right.full_id()] = {
            "shape": random.choice(
                [PulseShapeType.SOFT_SQUARE, PulseShapeType.SOFTER_SQUARE]
            ),
            "width": random.Random().uniform(100e-9, 250e-9),
            "rise": random.Random().uniform(1e-9, 100e-9),
            "amp": random.Random().uniform(1e6, 5e6),
            "drag": random.Random().uniform(0, 0.5),
            "phase": random.Random().uniform(0, 1),
        }
        hw.qubit_direction_couplings.append(
            QubitCoupling(
                direction=connection, quality=quality or random.Random().uniform(1, 100)
            )
        )

    hw.add_quantum_device(*qubit_devices, *resonator_devices)
    hw.is_calibrated = True
    return hw


def generate_random_linear(qubit_indices):
    output = {}
    for index in qubit_indices:
        random_0 = random.random()
        random_1 = random.random()
        output[index] = CalibratableUnitInterval2x2Array(
            [[random_0, 1 - random_1], [1 - random_0, random_1]]
        )
    return output


def check_type_legacy_or_pydantic(hw_model: QuantumHardwareModel | PydHardwareModel):
    if not isinstance(hw_model, QuantumHardwareModel | PydHardwareModel | None):
        raise TypeError(
            f"Invalid type for the hardware model: {hw_model.__class__.__name__}. Please provide a `QuantumHardwareModel` or `PhysicalHardwareModel`."
        )

    return hw_model


def hash_calibration_file(file_path: str | Path, algorithm="md5", chunk_size=8_192) -> str:
    """Compute the hash of a calibration file."""
    file_path = Path(file_path)

    hash_func = hashlib.new(algorithm)
    with file_path.open("rb") as file:
        # Byte chunks from the file and update hash.
        while chunk := file.read(chunk_size):
            hash_func.update(chunk)

    return hash_func.hexdigest()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from .compile import CompileSweepPipeline
from .execute import ExecuteSweepPipeline
from .passes import FrequencyAssignSanitisation

__all__ = [
    "CompileSweepPipeline",
    "ExecuteSweepPipeline",
    "FrequencyAssignSanitisation",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import abc
from contextlib import contextmanager

from qat.core.pass_base import PassManager
from qat.engines import ConnectionMixin, NativeEngine
from qat.model.target_data import TargetData
from qat.runtime.connection import ConnectionMode
from qat.runtime.passes.transform import (
    AcquisitionPostprocessing,
    AssignResultsTransform,
    InlineResultsProcessingTransform,
)


class BaseRuntime(abc.ABC):
    """Provides a Base class to build on for runtimes of varying complexities.

    A runtime provides the means to execute quantum programs. It can take on various
    responsibilities, including interfacing the execution engine and post-processing of
    results. Runtimes are designed to fit a specific purpose. For example, the
    :class:`SimpleRuntime` provides the means to execute already compiled programs
    :class:`Executable`. In the future, there will be support for hybrid runtimes that take
    on both compilation and execution responsibilities.
    """

    def __init__(
        self,
        engine: NativeEngine,
        results_pipeline: PassManager | None = None,
        connection_mode: ConnectionMode = ConnectionMode.DEFAULT,
    ):
        """
        :param engine: The execution engine for a target machine.
        :param results_pipeline: Optionally provided a pipeline for results processing. If
            not provided, a default pipeline is provided.
        :param connection_mode: Specifies how the connection is maintained.
        """
        self.engine = engine
        self.connection_mode = connection_mode
        self.connect_engine(ConnectionMode.CONNECT_AT_BEGINNING)

        if not results_pipeline:
            results_pipeline = self.default_pipeline()
        self.results_pipeline = results_pipeline

    @abc.abstractmethod
    def execute(package, *args, **kwargs): ...

    def default_pipeline(self, target_data: TargetData | None = None):
        return (
            PassManager()
            | AcquisitionPostprocessing(target_data)
            | InlineResultsProcessingTransform()
            | AssignResultsTransform()
        )

    def connect_engine(self, flag: ConnectionMode) -> bool | None:
        """Connect the engine according to the connection mode."""
        if not isinstance(self.engine, ConnectionMixin):
            return None

        if flag in self.connection_mode:
            if not self.engine.is_connected:
                self.engine.connect()
        return self.engine.is_connected

    def disconnect_engine(self, flag: ConnectionMode) -> bool | None:
        """Disconnect the engine according to the connection mode."""
        if not isinstance(self.engine, ConnectionMixin):
            return None

        if flag in self.connection_mode:
            if self.engine.is_connected:
                self.engine.disconnect()
        return self.engine.is_connected

    @contextmanager
    def _hold_connection(self):
        """Context manager to establish a connection which is held for the duration of the
        execution, according to the connection mode provided.

        If required, the engine will always be disconnected at the end of the context,
        even if execution is interrupted by an exception.
        """

        self.connect_engine(ConnectionMode.CONNECT_BEFORE_EXECUTE)
        try:
            yield self
        finally:
            self.disconnect_engine(ConnectionMode.DISCONNECT_AFTER_EXECUTE)

    def __del__(self):
        self.disconnect_engine(ConnectionMode.DISCONNECT_AT_END)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.disconnect_engine(ConnectionMode.DISCONNECT_AT_END)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
from copy import deepcopy
from numbers import Number
from typing import List, Union

import numpy as np
from compiler_config.config import InlineResultsProcessing, TketOptimizations
from pytket import Bit, Circuit, OpType, Qubit
from pytket._tket.architecture import Architecture, RingArch
from pytket._tket.circuit import CustomGateDef
from pytket._tket.predicates import (
    ConnectivityPredicate,
    DirectednessPredicate,
    MaxNQubitsPredicate,
    NoMidMeasurePredicate,
)
from pytket._tket.transform import Transform
from pytket.circuit import Node
from pytket.passes import (
    AutoRebase,
    CliffordSimp,
    ContextSimp,
    DecomposeArbitrarilyControlledGates,
    DecomposeBoxes,
    DefaultMappingPass,
    FullPeepholeOptimise,
    GlobalisePhasedX,
    KAKDecomposition,
    PeepholeOptimise2Q,
    RemoveBarriers,
    RemoveDiscarded,
    RemoveRedundancies,
    SequencePass,
    SimplifyMeasured,
    SynthesiseTket,
    ThreeQubitSquash,
)
from pytket.placement import Placement
from pytket.qasm import circuit_to_qasm_str
from pytket.qasm.qasm import NOPARAM_COMMANDS, PARAM_COMMANDS, QASMUnsupportedError
from qiskit.circuit.library import CXGate, ECRGate, UGate
from sympy import pi, sympify

from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.execution import InstructionExecutionEngine
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import Variable
from qat.purr.integrations.qasm import BitRegister, Qasm2Parser, QasmContext, QubitRegister
from qat.purr.integrations.qir import QIRParser
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class TketBuilder:
    """
    Builds a Tket circuit using a builder API. No direct relation to our other builders
    as the API is a little too different.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.circuit = Circuit()

    def qreg(self, name, size):
        self.circuit.add_q_register(name, size)

    def creg(self, name, size):
        self.circuit.add_c_register(name, size)

    def barrier(self, qubits):
        qubits = [qubits] if not isinstance(qubits, List) else qubits
        self.circuit.add_barrier(qubits)

    def ECR(self, qubit1, qubit2, *args):
        self.circuit.ECR(qubit1, qubit2)

    def measure(self, qubits, bits, conditions=None):
        qubits = [qubits] if not isinstance(qubits, List) else qubits
        bits = [bits] if not isinstance(bits, List) else bits
        conditions = conditions or {}
        for qubit, bit in zip(qubits, bits):
            self.circuit.Measure(qubit, bit, **conditions)

    def is_basic_gate(self, name):
        return self._get_basic_gate_def(name) is not None

    def _get_basic_gate_def(self, name):
        return PARAM_COMMANDS.get(name, None) or NOPARAM_COMMANDS.get(name, None)

    def gate(self, name, qubits, params=None, conditions=None):
        target_gate = self._get_basic_gate_def(name)
        if target_gate is None:
            raise ValueError(f"Can't find a definition for {name}.")

        conditions = conditions or {}
        params = params or []
        qubits = [qubits] if not isinstance(qubits, List) else qubits
        params = [params] if not isinstance(params, List) else params
        self.circuit.add_gate(target_gate, params, qubits, **conditions)

    def custom_gate(self, gate_def: CustomGateDef, qubits, params=None, conditions=None):
        conditions = conditions or {}
        params = params or []
        qubits = [qubits] if not isinstance(qubits, List) else qubits
        params = [params] if not isinstance(params, List) else params
        self.circuit.add_custom_gate(gate_def, params, qubits, **conditions)


class TketQasmParser(Qasm2Parser):
    """
    QASM parser that turns QASM into Tket structures. Switch to Tkets QASM parser when it's
    more mature.
    """

    def process_qreg(self, node, context: QasmContext, builder: TketBuilder, **kwargs):
        if node.name in context.registers.quantum:
            return
        context.registers.quantum[node.name] = QubitRegister(
            [Qubit(node.name, val) for val in range(node.size)]
        )
        builder.qreg(node.name, node.size)

    def process_creg(self, node, context, builder: TketBuilder, **kwargs):
        if node.name in context.registers.classic:
            return
        context.registers.classic[node.name] = BitRegister(
            [Bit(node.name, val) for val in range(node.size)]
        )
        builder.creg(node.name, node.size)

    def process_gate_definition(self, node, context, _, **kwargs):
        # Just a declaration, gates are used when CustomUnitary is used
        context.gates[node.name] = node

    def _get_staggered_qubit_params(self, node, context):
        def transform_to_halfturn(num):
            try:
                return sympify(num) / pi
            except Exception:
                raise ValueError("Cannot parse angle: {}".format(num))

        res = self._expand_to_match_registers(
            self._get_qubits(node, context),
            [
                transform_to_halfturn(param) if isinstance(param, Number) else param
                for param in self._get_parameters(node, context)
            ],
        )

        return res

    def process_gate(
        self,
        method,
        context: QasmContext,
        builder: TketBuilder,
        conditions=None,
        **kwargs,
    ):
        # For ease-of-use we just process CX and U as a normal gate, but the QASM nodes
        # don't have names. Infer them in this case.
        if isinstance(method.op, UGate):
            gate_name = "U"
        elif isinstance(method.op, CXGate):
            gate_name = "CX"
        elif isinstance(method.op, ECRGate):
            gate_name = "ECR"
        else:
            gate_name = method.name

        # TODO: Right now we just flatten all gates. Should move to use the
        #   MethodGateDef of Tket, but as that functionality is currently broken we
        #   lose nothing here.
        for qbs, args in self._get_staggered_qubit_params(method, context):
            if builder.is_basic_gate(gate_name):
                builder.gate(gate_name, qbs, args, conditions)
            else:
                super().process_gate(
                    method, context, builder, conditions=conditions, **kwargs
                )

    def process_if(self, node, context, builder: TketBuilder, *kwargs):
        left = self._resolve_variable(node.children[0], context)
        right = self._resolve_value(node.children[1], context)

        if isinstance(left, BitRegister):
            left = left.bits
        else:
            left = [left]

        conditions = {"condition_bits": left, "condition_value": right}
        self.walk_node(node.children[2], context, builder, conditions=conditions)

    def process_reset(self, node, context, builder, conditions=None, **kwargs):
        for qubits, _ in self._get_staggered_qubit_params(node, context):
            builder.gate("reset", qubits, conditions)

    def process_measure(
        self,
        node,
        context: QasmContext,
        builder: TketBuilder,
        conditions=None,
        **kwargs,
    ):
        for qubits, bits in self._get_staggered_qubit_params(node, context):
            builder.measure(qubits, bits, conditions)

    def process_unitary(
        self, node, context, builder: TketBuilder, conditions=None, **kwargs
    ):
        self.process_gate(node, context, builder, conditions, **kwargs)

    def process_cnot(self, node, context, builder, conditions=None, **kwargs):
        self.process_gate(node, context, builder, conditions, **kwargs)

    def process_barrier(self, node, context, builder: TketBuilder, **kwargs):
        barrier_qubits = []
        for qubits, _ in self._get_staggered_qubit_params(node, context):
            if isinstance(qubits, List):
                barrier_qubits.extend(qubits)
            else:
                barrier_qubits.append(qubits)

        builder.barrier(barrier_qubits)

    def process_program(self, builder, qasm):
        self._walk_program(builder, qasm, QasmContext())
        return builder


class TketQIRParser(QIRParser):
    """QIR parser than turns circuits into Tket structures."""

    def __init__(self, hardware: Union[QuantumHardwareModel, InstructionExecutionEngine]):
        if isinstance(hardware, InstructionExecutionEngine):
            hardware = hardware.model

        self.hardware: QuantumHardwareModel = hardware
        self.circuit: Circuit = Circuit(len(hardware.qubits), len(hardware.qubits))
        self.results_format = InlineResultsProcessing.Program
        self.result_variables = []

    def _get_qubit(self, id_: int):
        return self.hardware.qubits[id_]

    @staticmethod
    def normalize_parameter(param):
        return param / np.pi

    def ccx(self, control1, control2, target):
        self.circuit.CCX(control1, control2, target)

    def cx(self, control: int, target: int):
        self.circuit.CX(control, target)

    def cz(self, control: int, target: int):
        self.circuit.CZ(control, target)

    def h(self, target: int):
        self.circuit.H(target)

    def mz(self, qubit: int, target: int):
        self.circuit.Measure(qubit, target)

    def reset(self, target: int):
        self.circuit.Reset(target)

    def rx(self, theta: float, qubit: int):
        self.circuit.Rx(self.normalize_parameter(theta), qubit)

    def ry(self, theta: float, qubit: int):
        self.circuit.Ry(self.normalize_parameter(theta), qubit)

    def rz(self, theta: float, qubit: int):
        self.circuit.Rz(self.normalize_parameter(theta), qubit)

    def s(self, qubit: int):
        self.circuit.S(qubit)

    def s_adj(self, qubit: int):
        self.circuit.Sdg(qubit)

    def t(self, qubit: int):
        self.circuit.T(qubit)

    def t_adj(self, qubit: int):
        self.circuit.Tdg(qubit)

    def x(self, qubit: int):
        self.circuit.X(qubit)

    def y(self, qubit: int):
        self.circuit.Y(qubit)

    def z(self, qubit: int):
        self.circuit.Z(qubit)

    def returns(self, result_name=None):
        """Returns are dealt with upon converting back to QatIR."""
        pass

    def assign(self, name, value):
        """Assigns are dealt with upon converting back to QatIR."""
        pass

    @property
    def builder(self):
        """Returns the circuit and variables for where results are stored in Qat IR.

        This property overwrites the "builder" in the :class:`QIRParser` so that the circuit
        and output variables are returned in its place. It also removes unused qubits which
        can greatly increase the effectiveness of the placement of logical-to-physical
        qubits.
        """

        self.circuit.remove_blank_wires()
        return self.circuit, self.result_variables

    @builder.setter
    def builder(self, _):
        """The QIR parser resets the builder after being parsed (possibly for reuse). This
        setter is just to reset."""

        self.circuit = Circuit(len(self.hardware.qubits), len(self.hardware.qubits))


class TketToQatIRConverter:
    """Converts a Tket circuit into Qat IR.

    .. warning::
        This converter is only intended to be used to convert a TKET circuit into QAT IR
        after being parsed from QIR. It does not account for multiple quantum and classical
        registers, and might give undesired behaviour if used outside of this use case.
    """

    def __init__(self, model: QuantumHardwareModel):
        self.model = model
        self.builder = model.create_builder()
        self.output_variables = []

    def get_qubit(self, index: int):
        """Maps a Tket logical qubit index onto a physical qubit."""
        return self.model.qubits[index]

    @staticmethod
    def convert_parameter(arg: str):
        r"""A parameter stored in a Tket operation is in units of :math:`\pi`. Parameters
        are returned as a string, sometimes in fractional form: we need to convert it to an
        absolute value."""

        return float(pi * sympify(arg))

    def convert(self, circuit: Circuit):
        """Converts a Tket circuit into Qat IR, adding any necesarry assigns and returns.

        :param circuit: Program as a Tket circuit.
        :param result_format: Specifies how measurement results are formatted.
        """

        for command in circuit.to_dict()["commands"]:
            # Retrieves the qubit / clbit indices for each operation
            args = [arg[1][0] for arg in command["args"]]

            match command["op"]["type"]:
                # One-qubit gates
                case "X":
                    self.builder.X(self.get_qubit(args[0]))
                case "Y":
                    self.builder.Y(self.get_qubit(args[0]))
                case "Z":
                    self.builder.Z(self.get_qubit(args[0]))
                case "H":
                    self.builder.had(self.get_qubit(args[0]))
                case "SX":
                    self.builder.SX(self.get_qubit(args[0]))
                case "SXdg":
                    self.builder.SXdg(self.get_qubit(args[0]))
                case "S":
                    self.builder.S(self.get_qubit(args[0]))
                case "Sdg":
                    self.builder.Sdg(self.get_qubit(args[0]))
                case "T":
                    self.builder.T(self.get_qubit(args[0]))
                case "Tdg":
                    self.builder.Tdg(self.get_qubit(args[0]))
                case "Rx":
                    self.builder.X(
                        self.get_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                    )
                case "Ry":
                    self.builder.Y(
                        self.get_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                    )
                case "Rz":
                    self.builder.Z(
                        self.get_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                    )
                case "U1":
                    self.builder.Z(
                        self.get_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                    )
                case "U2":
                    self.builder.U(
                        self.get_qubit(args[0]),
                        np.pi / 2,
                        self.convert_parameter(command["op"]["params"][0]),
                        self.convert_parameter(command["op"]["params"][1]),
                    )
                case "U3":
                    self.builder.U(
                        self.get_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                        self.convert_parameter(command["op"]["params"][1]),
                        self.convert_parameter(command["op"]["params"][2]),
                    )

                # Two-qubit gates
                case "CX":
                    self.builder.cnot(self.get_qubit(args[0]), self.get_qubit(args[1]))
                case "ECR":
                    self.builder.ECR(self.get_qubit(args[0]), self.get_qubit(args[1]))
                case "SWAP":
                    self.builder.swap(self.get_qubit(args[0]), self.get_qubit(args[1]))

                # Operations
                case "Measure":
                    output_var = str(args[1])
                    self.builder.measure_single_shot_z(
                        self.get_qubit(args[0]), output_variable=output_var
                    )
                    self.output_variables.append(output_var)
                case "Barrier":
                    self.builder.synchronize([self.get_qubit(arg) for arg in args])
                case "Reset":
                    # Reset operation not implemented: do nothing instead of throwing an
                    # error to maintain support with non-optimised QIR.
                    continue
                case _:
                    raise NotImplementedError(
                        f"Command {command['op']['type']} not implemented."
                    )

        builder = self.builder
        self.builder = self.model.create_builder()
        return builder


def fetch_default_passes(architecture, opts, pass_list: List = None, add_delay=True):
    pass_list = pass_list or []
    if TketOptimizations.DefaultMappingPass in opts:
        pass_list.append(DefaultMappingPass(architecture, add_delay))

    return pass_list


def _full_stopalize(e):
    exception = str(e)
    return exception if exception.endswith(".") else f"{exception}."


def apply_default_transforms(circuit, architecture, opts):
    transform_list = []
    if TketOptimizations.DirectionalCXGates in opts:
        transform_list.append(Transform.DecomposeCXDirected(architecture))
        transform_list.append(Transform.RemoveRedundancies())

    for transform in transform_list:
        transform.apply(circuit)


def check_validity(circuit, architecture):
    predicates = [
        (
            "Failed to apply circuit to hardware topology.",
            ConnectivityPredicate(architecture),
        ),
        (
            "Failed to satisfy 2QB gate constraints.",
            DirectednessPredicate(architecture),
        ),
        ("No mid-circuit measurements allowed.", NoMidMeasurePredicate()),
    ]

    failed_preds = [message for message, pred in predicates if not pred.verify(circuit)]
    if any(failed_preds):
        raise ValueError(f"Verification failed: {' '.join(failed_preds)}")


def optimize_circuit(circ, architecture, opts):
    """
    Apply all Tket optimizations to the circuit provided as defined in the options flag.
    """
    try:
        # By default we want to decompose Tket-specific circuits and then rebase into
        # the form that Tket can optimize most efficiently.
        passes = [DecomposeBoxes(), SynthesiseTket()]

        if TketOptimizations.ContextSimp in opts:
            passes.append(ContextSimp(allow_classical=True))

        # TODO: FullPeepholeOptimise fails due to bug in tket. Find out what's wrong and
        #   fix/flag.
        if TketOptimizations.FullPeepholeOptimise in opts:
            passes.append(FullPeepholeOptimise())

        if TketOptimizations.CliffordSimp in opts:
            passes.append(CliffordSimp(allow_swaps=True))

        if TketOptimizations.DecomposeArbitrarilyControlledGates in opts:
            passes.append(DecomposeArbitrarilyControlledGates())

        # if TketOptimizations.EulerAngleReduction in opts:
        #     passes.append(EulerAngleReduction())

        if TketOptimizations.GlobalisePhasedX in opts:
            passes.append(GlobalisePhasedX())

        # if TketOptimizations.GuidedPauliSimp in opts:
        #     passes.append(GuidedPauliSimp())

        if TketOptimizations.KAKDecomposition in opts:
            passes.append(KAKDecomposition())

        # if TketOptimizations.OptimisePhaseGadgets in opts:
        #     passes.append(OptimisePhaseGadgets())
        #
        # if TketOptimizations.PauliSimp in opts:
        #     passes.append(PauliSimp())
        #
        # if TketOptimizations.PauliSquash in opts:
        #     passes.append(PauliSquash())

        if TketOptimizations.PeepholeOptimise2Q in opts:
            passes.append(PeepholeOptimise2Q())

        if TketOptimizations.RemoveDiscarded in opts:
            passes.append(RemoveDiscarded())

        if TketOptimizations.RemoveBarriers in opts:
            passes.append(RemoveBarriers())

        if TketOptimizations.RemoveRedundancies in opts:
            passes.append(RemoveRedundancies())

        if TketOptimizations.ThreeQubitSquash in opts:
            passes.append(ThreeQubitSquash())

        if TketOptimizations.SimplifyMeasured in opts:
            passes.append(SimplifyMeasured())

        # TODO: Use HW model object to create fetch calibration data for optimal
        #   routing.
        passes = fetch_default_passes(architecture, opts, passes)

        passes += [AutoRebase({OpType.CX, OpType.U3})]

        # Not everything in the list is a pass, we've also got transforms.
        # Everything in the list should have an apply function though.
        SequencePass(passes, strict=False).apply(circ)
        apply_default_transforms(circ, architecture, opts)
        check_validity(circ, architecture)

        log.info("Tket optimizations complete.")
        return True
    except (ValueError, IndexError, RuntimeError) as e:
        log.warning(
            f"Tket could not compile circuit due do exception: {_full_stopalize(e)}. "
            "Skipping this optimization pass."
        )
        return False


def get_coupling_subgraphs(couplings):
    """
    Given a list of connections which potentially describe a disconnected graph,
    this returns a list of connected subgraphs
    """
    subgraphs = []
    for coupling in couplings:
        subgraph_added_to = None
        for subgraph in subgraphs:
            # check whether coupling joins to subgraph
            if set(coupling) & {qubit for connection in subgraph for qubit in connection}:
                if subgraph_added_to is None:
                    subgraph.append(coupling)  # add the coupling to the subgraph
                    subgraph_added_to = subgraph
                else:
                    subgraph_added_to.extend(
                        subgraph
                    )  # if coupling joins to 2nd subgraph, the subgraphs combine
                    subgraphs.remove(subgraph)
        if (
            subgraph_added_to is None
        ):  # if coupling doesn't join to any of the current subgraphs, make it a new one
            subgraphs.append([coupling])
    return subgraphs


def run_1Q_tket_optimizations(circ: Circuit, hardware: QuantumHardwareModel) -> Circuit:
    logical_qubit_map = {q.index: i for i, q in enumerate(hardware.qubits)}

    qubit_qualities = {}
    for phys_q, log_q in logical_qubit_map.items():
        qubit_qualities[log_q] = (
            hardware.qubit_quality(phys_q) if hardware.qubit_quality(phys_q) != 1 else 0
        )
    best_qubit = max(qubit_qualities, key=qubit_qualities.get)

    q_map = {Qubit(0): Node(best_qubit), Qubit(best_qubit): Node(0)}
    Placement.place_with_map(circ, q_map)

    return circ


def run_multiQ_tket_optimizations(
    circ: Circuit, opts, hardware: QuantumHardwareModel, use_1Q_quality: bool = False
) -> Circuit:
    couplings = deepcopy(hardware.qubit_direction_couplings)
    optimizations_failed = False
    architecture = None
    if any(couplings):
        if TketOptimizations.DefaultMappingPass not in opts:
            architecture = Architecture([val.direction for val in couplings])
            optimizations_failed = not optimize_circuit(circ, architecture, opts)

        else:
            # Without default remapping pass multi-qubit gates don't get moved around, so
            # trying to apply them to a limited subset of qubits provides no value.
            logical_qubit_map = {q.index: i for i, q in enumerate(hardware.qubits)}
            for coupling in couplings:
                control_ind, target_ind = coupling.direction
                coupling.direction = (
                    logical_qubit_map[control_ind],
                    logical_qubit_map[target_ind],
                )
                if use_1Q_quality:
                    coupling.quality *= hardware.qubit_quality(
                        control_ind
                    ) * hardware.qubit_quality(target_ind)
            else:
                coupling_qualities = list({val.quality for val in couplings})
            coupling_qualities.sort(reverse=True)

            for quality_level in coupling_qualities:
                filtered_couplings = [
                    val.direction for val in couplings if val.quality >= quality_level
                ]
                coupling_subgraphs = get_coupling_subgraphs(filtered_couplings)

                for subgraph in coupling_subgraphs:
                    if circ.n_qubits <= len(
                        set([qubit for coupling in subgraph for qubit in coupling])
                    ):
                        architecture = Architecture(subgraph)
                        optimizations_failed = not optimize_circuit(
                            circ, architecture, opts
                        )
                        if not optimizations_failed:
                            break
                    else:
                        optimizations_failed = True
                if not optimizations_failed:
                    break
    else:
        architecture = RingArch(len(hardware.qubits))
        optimizations_failed = not optimize_circuit(circ, architecture, opts)

    # If our optimizations failed but we want the mapping pass, apply that by itself.
    if optimizations_failed:
        if architecture is None:
            raise ValueError(
                "Unable to resolve hardware instance for fall-back optimizations."
            )

        delay_failed = False
        try:
            # DelayMeasure throws on failure, and we want to raise our own errors for
            # this.
            SequencePass(fetch_default_passes(architecture, opts)).apply(circ)
        except RuntimeError:
            delay_failed = True

        # If the delay fails, try with a more limited subset.
        if delay_failed:
            try:
                # Tket just throws an exception if the list is none, so skip if that's
                # the case.
                default_passes = fetch_default_passes(architecture, opts, add_delay=False)
                if len(default_passes) > 0:
                    SequencePass(default_passes).apply(circ)
            except RuntimeError as e:
                message = str(e)
                if MaxNQubitsPredicate.__name__ in message:
                    raise ValueError(
                        f"Circuit uses {len(circ.qubits)} qubits, "
                        f"only {len(architecture.nodes)} available."
                    )

                raise e

        apply_default_transforms(circ, architecture, opts)
        check_validity(circ, architecture)
    return circ


def run_tket_optimizations(circ: Circuit, opts, hardware: QuantumHardwareModel) -> Circuit:
    """
    Runs tket-based optimizations and modifications. Routing will always happen no
    matter the level.

    Will run optimizations in sections if a full suite fails until a minimal subset of
    passing optimizations is found.

    :param circ: Tket circuit to optimize. The source file must be already parsed to a
        TKET circuit.
    :param opts: Specifies which TKET optimizations to run.
    :param hardware: The hardware model is used for routing and placement purposes.
    """
    if circ.n_qubits == 1 and hardware.error_mitigation:
        return run_1Q_tket_optimizations(circ, hardware)
    elif circ.n_qubits == 2 and hardware.error_mitigation:
        return run_multiQ_tket_optimizations(circ, opts, hardware, use_1Q_quality=True)
    else:
        return run_multiQ_tket_optimizations(circ, opts, hardware, use_1Q_quality=False)


def run_tket_optimizations_qasm(qasm_string, opts, hardware: QuantumHardwareModel) -> str:
    """
    Runs tket-based optimizations and modifications. Routing will always happen no
    matter the level.

    Will run optimizations in sections if a full suite fails until a minimal subset of
    passing optimizations is found.

    :param qasm_string: The circuit as a QASM string.
    :param opts: Specifies which TKET optimizations to run.
    :param hardware: The hardware model is used for routing and placement purposes.
    """
    try:
        tket_builder: TketBuilder = TketQasmParser().parse(TketBuilder(), qasm_string)
        circ = tket_builder.circuit
        log.info(f"Number of gates before tket optimization: {circ.n_gates}")
    except Exception as e:  # Parsing is too fragile, can cause almost any exception.
        log.warning(
            f"Tket failed during QASM parsing with error: {_full_stopalize(e)}. "
            "Skipping this optimization pass."
        )
        return qasm_string

    circ = run_tket_optimizations(circ, opts, hardware)

    try:
        qasm_string = circuit_to_qasm_str(circ)
        log.info(f"Number of gates after tket optimization: {circ.n_gates}")
    except (QASMUnsupportedError, RuntimeError) as e:
        log.warning(
            f"Error generating QASM from Tket circuit: {_full_stopalize(e)}. "
            "Skipping this optimization pass."
        )

    # TODO: Return result object with more information about compilation/errors
    return qasm_string


def run_tket_optimizations_qir(
    file_or_str,
    opts,
    hardware: QuantumHardwareModel,
    results_format: InlineResultsProcessing = None,
) -> InstructionBuilder:
    """
    Runs tket-based optimizations and modifications. Routing will always happen no
    matter the level.

    Will run optimizations in sections if a full suite fails until a minimal subset of
    passing optimizations is found.

    :param file_or_str: The QIR program as a string, or its file path.
    :param opts: Specifies which TKET optimizations to run.
    :param hardware: The hardware model is used for routing and placement purposes.
    """
    results_format = results_format or InlineResultsProcessing.Program
    circ, output_variables = TketQIRParser(hardware).parse(file_or_str)
    log.info(f"Number of gates before tket optimization: {circ.n_gates}")

    circ = run_tket_optimizations(circ, opts, hardware)

    builder = TketToQatIRConverter(hardware).convert(circ)
    for output_variable in output_variables:
        builder.results_processing(output_variable[0].name, results_format)
    if any(output_variables):
        potential_names = [val[1] for val in output_variables if len(val[1] or "") != 0]
        if not any(potential_names):
            result_name = Variable.generate_name()
        else:
            result_name = "_".join(potential_names)
        builder.assign(result_name, [val[0] for val in output_variables])
        builder.returns(result_name)
    else:
        builder.returns()
    log.info(f"Number of gates after tket optimization: {circ.n_gates}")
    return builder



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd


from qat.core.pass_base import PassManager
from qat.middleend.base import CustomMiddleend
from qat.middleend.passes.analysis import ActivePulseChannelAnalysis
from qat.middleend.passes.transform import (
    BatchedShots,
    EndOfTaskResetSanitisation,
    EvaluateWaveforms,
    FreqShiftSanitisation,
    InactivePulseChannelSanitisation,
    InitialPhaseResetSanitisation,
    InstructionGranularitySanitisation,
    InstructionLengthSanitisation,
    LowerSyncsToDelays,
    MeasurePhaseResetSanitisation,
    PhaseOptimisation,
    PopulateWaveformSampleTime,
    PostProcessingSanitisation,
    RepeatSanitisation,
    RepeatTranslation,
    ResetsToDelays,
    ReturnSanitisation,
    ScopeSanitisation,
    SquashDelaysOptimisation,
    SynchronizeTask,
)
from qat.middleend.passes.validation import (
    DynamicFrequencyValidation,
    FrequencySetupValidation,
    HardwareConfigValidity,
    ReadoutValidation,
)
from qat.model.hardware_model import PhysicalHardwareModel as HardwareModel
from qat.model.target_data import TargetData


class DefaultMiddleend(CustomMiddleend):
    """The standard middle end used for antic pipelines.

    Implements a number of passes, including validation and sanitisation of IR, and
    optimizations. Also implements a number of lowering passes, e.g., lowering repeats to
    more explicit control flow instructions.
    """

    def __init__(
        self,
        model: HardwareModel,
        target_data: TargetData = TargetData.default(),
    ):
        """
        :param model: The hardware model that holds calibrated information on the qubits on
            the QPU.
        :param clock_cycle: The period for a single sequencer clock cycle.
        """
        pipeline = self.build_pass_pipeline(model, target_data)
        self.target_data = target_data
        super().__init__(model=model, pipeline=pipeline)

    @staticmethod
    def build_pass_pipeline(
        model: HardwareModel,
        target_data: TargetData = TargetData.default(),
    ) -> PassManager:
        """
        Builds the default middle end pass pipeline.
        :param model: The hardware model that holds calibrated information on the qubits on
            the QPU.
        :return: A :class:`PassManager` containing a sequence of passes.
        """
        return (
            PassManager()
            | PopulateWaveformSampleTime(model, target_data)
            | HardwareConfigValidity(model)
            | ActivePulseChannelAnalysis(model)
            | InactivePulseChannelSanitisation()
            | FrequencySetupValidation(model, target_data)
            | DynamicFrequencyValidation(model, target_data)
            # Sanitising input IR to make it complete
            | RepeatSanitisation(target_data)
            | ReturnSanitisation()
            | SynchronizeTask()
            | PostProcessingSanitisation()
            | ReadoutValidation()
            | MeasurePhaseResetSanitisation(model)
            | InstructionGranularitySanitisation(target_data)
            # Preparing for codegen
            | EvaluateWaveforms(model, target_data)
            | LowerSyncsToDelays()
            | FreqShiftSanitisation(model)
            | InitialPhaseResetSanitisation()
            | PhaseOptimisation()
            | EndOfTaskResetSanitisation(model)
            | ResetsToDelays(model, target_data)
            | SquashDelaysOptimisation()
            | InstructionLengthSanitisation(model, target_data)
            | BatchedShots(target_data)
            | ScopeSanitisation()
            | RepeatTranslation(target_data)
        )


PydDefaultMiddleend = DefaultMiddleend



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd


class BasicBlock:
    def __init__(self, elements=None):
        self.elements: list = elements or []

    def head(self):
        if self.elements:
            return self.elements[0]
        return None

    def tail(self):
        if self.elements:
            return self.elements[-1]
        return None

    def is_empty(self):
        return len(self.elements) == 0

    def iterator(self):
        return iter(self.elements)

    def __repr__(self):
        return self.elements


class Flow:
    def __init__(self, src=None, dest=None):
        self.src: BasicBlock = src
        self.dest: BasicBlock = dest

    def __repr__(self):
        return (
            f"{self.src.head()} .. {self.src.tail()}",
            f"{self.dest.head()} .. {self.dest.tail()}",
        )


class ControlFlowGraph:
    def __init__(self, nodes=None, edges=None):
        self.nodes: list[BasicBlock] = nodes or []
        self.edges: list[Flow] = edges or []
        self.entry = None

    def get_or_create_node(self, header: int) -> BasicBlock:
        node = next((n for n in self.nodes if n.head() == header), None)
        if not node:
            node = BasicBlock([header])
            self.nodes.append(node)

        self.entry = self.entry or node
        return node

    def get_or_create_edge(self, src: BasicBlock, dest: BasicBlock) -> Flow:
        edge = next((f for f in self.edges if f.src == src and f.dest == dest), None)
        if not edge:
            edge = Flow(src, dest)
            self.edges.append(edge)
        return edge

    def out_nbrs(self, node) -> list[BasicBlock]:
        return [e.dest for e in self.edges if e.src == node]

    def in_nbrs(self, node) -> list[BasicBlock]:
        return [e.src for e in self.edges if e.dest == node]

    def out_edges(self, node) -> list[Flow]:
        return [e for e in self.edges if e.src == node]

    def in_edges(self, node) -> list[Flow]:
        return [e for e in self.edges if e.dest == node]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from abc import ABC
from typing import Dict

from pydantic import BaseModel, Field, IPvAnyAddress

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class InstrumentModel(BaseModel):
    """
    Used to parse JSON/CSV entries. An instrument has an id, name, and IP address
    """

    id: str = Field(frozen=True, alias="ID")
    name: str = Field(frozen=True, alias="NAME")
    address: IPvAnyAddress = Field(frozen=True, alias="ADDRESS")


class InstrumentConcept(ABC):
    """
    Basic APIs through which an instrument must be able to interact with
    """

    def connect(self):
        pass

    def disconnect(self):
        pass

    def close(self):
        pass

    def upload(self, *args, **kwargs):
        pass

    def playback(self, *args, **kwargs):
        pass

    def collect(self, *args, **kwargs):
        pass


class LeafInstrument(InstrumentConcept):
    def __init__(self, parse_model: InstrumentModel):
        self.id: str = parse_model.id
        self.name: str = parse_model.name
        self.address: str = str(parse_model.address)

        self.is_connected = False

    def __repr__(self):
        return f"{self.id}_{self.name}_{self.address}"

    def __str__(self):
        return self.__repr__()


class CompositeInstrument(InstrumentConcept):
    def __init__(self):
        self.components: Dict[str, LeafInstrument] = {}

    def connect(self):
        for comp in self.components.values():
            comp.connect()

    def disconnect(self):
        for comp in self.components.values():
            comp.disconnect()

    def close(self):
        for comp in self.components.values():
            comp.close()

    def add(self, component: LeafInstrument):
        comp = next(
            (comp for comp in self.components.values() if comp.id == component.id),
            None,
        )
        if comp:
            log.warning(f"Instrument {component} already exists")
        else:
            self.components[component.id] = component

    def remove(self, component: LeafInstrument):
        if component.id in self.components:
            del self.components[component.id]

    def lookup_by_name(self, name: str) -> LeafInstrument:
        comp = next((comp for comp in self.components.values() if comp.name == name), None)
        if not comp:
            raise ValueError(f"Could not find instrument with name {name}")
        return comp



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd


class MismatchingHardwareModelException(Exception):
    """Raised when the hardware model does not match in the compile and execute modes."""

    ...



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.backend.qblox.codegen import QbloxBackend2
from qat.backend.qblox.config.constants import QbloxTargetData
from qat.core.pass_base import PassManager
from qat.frontend import AutoFrontend
from qat.middleend import CustomMiddleend
from qat.middleend.passes.purr.transform import (
    DeviceUpdateSanitisation,
    PhaseOptimisation,
    PostProcessingSanitisation,
)
from qat.middleend.passes.purr.validation import InstructionValidation, ReadoutValidation
from qat.pipelines.pipeline import CompilePipeline
from qat.pipelines.purr.qblox.compile import backend_pipeline2
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def middleend_pipeline(
    model: QuantumHardwareModel, target_data: QbloxTargetData
) -> PassManager:
    """Factory for creating middleend pipelines for Qblox legacy models.

    Includes a list of passes that replicate the responsibilities of
    :meth:`LiveDeviceEngine.optimize <qat.purr.backends.live.LiveDeviceEngine.optimize>` and
    :meth:`LiveDeviceEngine.validate <qat.purr.backends.live.LiveDeviceEngine.validate>`.

    :param model: The hardware model is required for ReadoutValidation.
    :return: The pipeline as a pass manager.
    """
    return (
        PassManager()
        | DeviceUpdateSanitisation()
        | PhaseOptimisation()
        | PostProcessingSanitisation()
        | InstructionValidation(target_data)
        | ReadoutValidation(model)
    )


class LegacyQbloxCompilePipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the legacy qblox backend.

    Implements a custom pipeline to make instructions suitable for the legacy qblox engine,
    and cannot be configured with a custom engine.

    .. warning::

        This pipeline is for compilation purposes only and does not execute programs. Please
        select an appropriate execution pipeline if you wish to run the compiled programs.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: QbloxTargetData = None,
        engine=None,
    ) -> CompilePipeline:
        if engine is not None:
            log.warning(
                "The compilation pipeline does not require an engine. It will be ignored."
            )

        target_data = target_data or QbloxTargetData.default()
        return CompilePipeline(
            name=config.name,
            model=model,
            target_data=target_data,
            frontend=AutoFrontend.default_for_legacy(model),
            middleend=CustomMiddleend(
                model,
                pipeline=middleend_pipeline(model=model, target_data=target_data),
            ),
            backend=QbloxBackend2(
                model,
                pipeline=backend_pipeline2(),
            ),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from typing import List

from qat.purr.backends.qblox.codegen import QbloxPackage
from qat.purr.backends.qblox.execution.instrument_base import (
    CompositeInstrument,
    LeafInstrument,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class LeafExecutor(LeafInstrument):
    def upload(self, packages: List[QbloxPackage]):
        pass

    def playback(self):
        pass

    def collect(self):
        pass


class CompositeExecutor(CompositeInstrument):
    def upload(self, packages: List[QbloxPackage]):
        pass

    def playback(self, *args, **kwargs):
        pass

    def collect(self, *args, **kwargs):
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.backend.fallthrough import FallthroughBackend
from qat.core.pass_base import PassManager
from qat.frontend import AutoFrontend
from qat.middleend import CustomMiddleend
from qat.middleend.passes.purr.transform import (
    LegacyPhaseOptimisation,
    PostProcessingSanitisation,
)
from qat.middleend.passes.purr.validation import (
    HardwareConfigValidity,
    InstructionValidation,
    ReadoutValidation,
)
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import CompilePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime.passes.purr.analysis import CalibrationAnalysis

log = get_default_logger()


def middleend_pipeline(model: QuantumHardwareModel, target_data: TargetData) -> PassManager:
    return (
        PassManager()
        | HardwareConfigValidity(model)
        | CalibrationAnalysis()
        | LegacyPhaseOptimisation()
        | PostProcessingSanitisation()
        | InstructionValidation(target_data)
        | ReadoutValidation(model)
    )


class LegacyCompilePipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the legacy backend.

    .. warning::

        This pipeline is for compilation purposes only and does not execute programs. Please
        select an appropriate execution pipeline if you wish to run the compiled programs.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None = None,
        engine: None = None,
    ) -> CompilePipeline:
        if engine is not None:
            log.warning(
                "The compilation pipeline does not require an engine. It will be ignored."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        return CompilePipeline(
            name=config.name,
            model=model,
            target_data=target_data,
            frontend=AutoFrontend.default_for_legacy(model),
            middleend=CustomMiddleend(
                model,
                pipeline=middleend_pipeline(model=model, target_data=target_data),
            ),
            backend=FallthroughBackend(model),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from dataclasses import dataclass, field

import numpy as np

from qat.core.pass_base import AnalysisPass
from qat.core.result_base import ResultInfoMixin, ResultManager
from qat.ir.instructions import QuantumInstruction
from qat.ir.lowered import PartitionedIR
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData


@dataclass
class PulseChannelTimeline:
    """Timeline analysis for instructions on a pulse channel.

    Imagine the timeline for a pulse channel, with an instruction that occurs over samples
    3-7, i.e.,

        samples: 0 1 2 [3 4 5 6 7] 8 9 10.

    The `start_position` would be 3, the `end_position` 7, and the number of `samples` 5.

    :param np.ndarray[int] samples: The number of samples each instruction takes.
    :param np.ndarray[int] start_positions: The sample when the instruction begins.
    :param np.ndarray[int] end_positions: The sample when the instruction ends.
    """

    samples: np.ndarray[int] = field(default_factory=lambda: np.ndarray([]))
    start_positions: np.ndarray[int] = field(default_factory=lambda: np.ndarray([]))
    end_positions: np.ndarray[int] = field(default_factory=lambda: np.ndarray([]))


@dataclass
class TimelineAnalysisResult(ResultInfoMixin):
    """Stores the timeline analysis for all pulse channels.

    :param target_map: The dictionary containing the timeline analysis for all pulse
        channels.
    """

    target_map: dict[str, PulseChannelTimeline] = field(
        default_factory=lambda: defaultdict(PulseChannelTimeline)
    )
    total_duration: float = field(default=0.0)


class TimelineAnalysis(AnalysisPass):
    """Analyses the timeline of each pulse channel.

    Takes the instruction list for each pulse channel retrieved from the the partitioned
    results, and calculates the timeline in units of samples (each sample takes time
    `sample_time`). It calculates the duration of each instruction in units of samples,
    and the start and end times of each instruction in units of samples.

    .. warning::

        The pass will assume that the durations of instructions are sanitised to the
        granularity of the channels. If instructions that do not meet the criteria are
        provided, it might produce incorrect timelines. This can be enforced used the
        :class:`InstructionGranularitySanitisation <qat.middleend.passes.transform.InstructionGranularitySanitisation>`
        pass.
    """

    def __init__(self, model: PhysicalHardwareModel, target_data: TargetData):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
        :param target_data: Target-related information.
        """
        self.sample_times_map = dict()
        for qubit in model.qubits.values():
            self.sample_times_map[qubit.physical_channel.uuid] = (
                target_data.QUBIT_DATA.sample_time
            )
            self.sample_times_map[qubit.resonator.physical_channel.uuid] = (
                target_data.RESONATOR_DATA.sample_time
            )

    def run(
        self, ir: PartitionedIR, res_mgr: ResultManager, *args, **kwargs
    ) -> PartitionedIR:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """
        target_map = ir.target_map

        result = TimelineAnalysisResult()
        total_duration = 0
        for pulse_ch_id, instructions in target_map.items():
            pulse_channel_durations = np.array(
                [
                    (inst.duration if isinstance(inst, QuantumInstruction) else 0.0)
                    for inst in instructions
                ]
            )
            total_duration = max(total_duration, np.sum(pulse_channel_durations))

            sample_time = self.sample_times_map[
                ir.get_pulse_channel(pulse_ch_id).physical_channel_id
            ]
            durations = self.durations_as_samples(pulse_channel_durations, sample_time)
            cumulative_durations = np.cumsum(durations)
            result.target_map[pulse_ch_id] = PulseChannelTimeline(
                samples=durations,
                start_positions=cumulative_durations - durations,
                end_positions=cumulative_durations - 1,
            )

        result.total_duration = total_duration
        res_mgr.add(result)
        return ir

    def durations_as_samples(self, durations: list[float], sample_time: float):
        """Converts a list of durations into a number of samples."""
        block_numbers = np.ceil(np.round(durations / sample_time, decimals=4)).astype(
            np.int64
        )
        return block_numbers


PydTimelineAnalysisResult = TimelineAnalysisResult
PydTimelineAnalysis = TimelineAnalysis



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.model.loaders.purr.base import BaseLegacyModelLoader
from qat.purr.backends.echo import Connectivity
from qat.purr.backends.qblox.dummy import get_default_dummy_hardware


class QbloxDummyModelLoader(BaseLegacyModelLoader):
    def __init__(
        self,
        name: str = None,
        dummy_cfg: dict = None,
        qubit_count: int = 4,
        connectivity: Connectivity | list[(int, int)] | None = None,
        add_direction_couplings=True,
    ):
        self.name = name
        self.dummy_cfg = dummy_cfg
        self.connectivity = connectivity
        self.qubit_count = qubit_count
        self.add_direction_couplings = add_direction_couplings

    def load(self):
        return get_default_dummy_hardware(
            qubit_count=self.qubit_count,
            connectivity=self.connectivity,
            add_direction_couplings=self.add_direction_couplings,
            name=self.name,
            dummy_cfg=self.dummy_cfg,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import itertools
from collections import OrderedDict, defaultdict
from copy import deepcopy
from functools import singledispatchmethod

import numpy as np
from compiler_config.config import CompilerConfig, MetricsType
from more_itertools import partition
from numpy.typing import NDArray

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import TransformPass
from qat.core.result_base import ResultManager
from qat.ir.instruction_builder import InstructionBuilder, QuantumInstructionBuilder
from qat.ir.instructions import (
    Assign,
    Delay,
    EndRepeat,
    GreaterThan,
    Instruction,
    Jump,
    Label,
    LoopCount,
    PhaseReset,
    PhaseSet,
    PhaseShift,
    Plus,
    QuantumInstruction,
    Repeat,
    Reset,
    Return,
    Synchronize,
    Variable,
)
from qat.ir.measure import Acquire, MeasureBlock, PostProcessing
from qat.ir.waveforms import Pulse, SampledWaveform, SquareWaveform, Waveform
from qat.middleend.passes.analysis import ActivePulseChannelResults
from qat.model.device import FreqShiftPulseChannel, PulseChannel, Qubit
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData
from qat.purr.compiler.instructions import AcquireMode, PostProcessType, ProcessAxis
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class PopulateWaveformSampleTime(TransformPass):
    """Populates instructions within the IR with the sample time from their coupled
    physical channel.

    This is a symptom of the IR builder not having access to the target data when assembling
    the instructions; this will likely be addressed later.
    """

    def __init__(self, hardware_model: PhysicalHardwareModel, target_data: TargetData):
        """
        :param hardware_model: The hardware model that holds calibrated information on the
            qubits on the QPU.
        :param target_data: Target-related information.
        """
        self._channel_data = self._create_channel_data(hardware_model, target_data)

    @staticmethod
    def _create_channel_data(
        hardware_model: PhysicalHardwareModel, target_data
    ) -> dict[str, float]:
        """Maps physical channels onto their sample times."""
        channel_data = {}

        for qubit in hardware_model.qubits.values():
            channel_data[qubit.physical_channel.uuid] = target_data.QUBIT_DATA.sample_time
            channel_data[qubit.resonator.physical_channel.uuid] = (
                target_data.RESONATOR_DATA.sample_time
            )
        return channel_data

    def run(
        self, ir: QuantumInstructionBuilder, *args, **kwargs
    ) -> QuantumInstructionBuilder:
        for inst in ir.instructions:
            target = getattr(inst, "target", None)
            if target is None:
                continue
            physical_channel_id = ir.get_pulse_channel(target).physical_channel_id
            self._add_sample_time(inst, physical_channel_id)
        return ir

    @singledispatchmethod
    def _add_sample_time(self, inst: Instruction, target: str):
        pass

    @_add_sample_time.register(Pulse)
    def _(self, inst: Pulse, target: str):
        self._add_sample_time(inst.waveform, target)
        inst.duration = inst.waveform.duration

    @_add_sample_time.register(Acquire)
    def _(self, inst: Acquire, target: str):
        if inst.filter is not None:
            self._add_sample_time(inst.filter, target)

    @_add_sample_time.register(SampledWaveform)
    def _(self, inst: SampledWaveform, target: str):
        inst.sample_time = self._channel_data[target]


class RepeatTranslation(TransformPass):
    """Transform :class:`Repeat` instructions so that they are replaced with:
    :class:`Variable`, :class:`Assign`, and :class:`Label` instructions at the start,
    and :class:`Assign` and :class:`Jump` instructions at the end."""

    def __init__(self, target_data: TargetData):
        self.target_data = target_data

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        handler = RepeatTranslationHandler()
        for inst in ir.instructions:
            handler.run(inst)

        handler.close_repeats()

        if getattr(ir, "passive_reset_time", None) is None:
            ir.passive_reset_time = self.target_data.QUBIT_DATA.passive_reset_time
        ir.instructions = handler.instructions
        return ir


class RepeatTranslationHandler:
    """
    Handler used by PydRepeatTranslation to manage the translation of Repeat and EndRepeat instructions.
    """

    def __init__(self):
        self.instructions: list[Instruction] = []
        self.label_data: OrderedDict[Label, dict[str, Repeat | Variable]] = OrderedDict()

    @singledispatchmethod
    def run(self, instruction):
        """
        Default handler for instructions that do not have a specific repeat translation.
        """
        raise NotImplementedError(f"No repeat translation for {type(instruction)}")

    @run.register(Instruction)
    def _(self, instruction: Instruction):
        self.instructions.append(instruction)

    @run.register(Repeat)
    def _(self, instruction: Repeat):
        label = Label.with_random_name()
        var = Variable(name=label.name + "_count", var_type=LoopCount)
        assign = Assign(name=var.name, value=0)

        self.instructions.extend([var, assign, label])
        self.label_data[label.name] = {"repeat": instruction, "var": var}

    @run.register(EndRepeat)
    def _(self, instruction: EndRepeat):
        if len(self.label_data) < 1:
            raise ValueError("EndRepeat found without associated Repeat.")
        label, data = self.label_data.popitem()
        self.instructions.extend(self._close_repeat(label, data))

    def close_repeats(self):
        for label, data in reversed(self.label_data.items()):
            self.instructions.extend(self._close_repeat(label, data))

    @staticmethod
    def _close_repeat(label: Label, data: dict):
        # TODO: Adjust so that closing is done before finishing instructions such as
        #   returns and postprocessing. COMPILER-452
        repeat = data["repeat"]
        var = data["var"]
        assign = Assign(name=var.name, value=Plus(left=var, right=1))
        jump = Jump(label=label, condition=GreaterThan(left=repeat.repeat_count, right=var))
        return [assign, jump]


class PhaseOptimisation(TransformPass):
    """Iterates through the list of instructions and compresses contiguous
    :class:`PhaseShift` instructions.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        :param met_mgr: The metrics manager to store the number of instructions after
            optimisation.
        """

        handler = PhaseOptimisationHandler()
        for instruction in ir:
            handler.run(instruction)

        ir.instructions = handler.optimized_instructions
        met_mgr.record_metric(
            MetricsType.OptimizedInstructionCount, ir.number_of_instructions
        )
        return ir


class PhaseOptimisationHandler:
    def __init__(self):
        self.accum_phaseshifts: dict[str, float] = defaultdict(float)
        self.optimized_instructions: list = []
        self.previous_phase_instruction_is_phase_set = defaultdict(bool)

    @singledispatchmethod
    def run(self, instruction):
        self.optimized_instructions.append(instruction)

    @run.register(PhaseShift)
    def _(self, instruction: PhaseShift):
        self.accum_phaseshifts[instruction.target] += instruction.phase

    @run.register(PhaseSet)
    def _(self, instruction: PhaseSet):
        self.accum_phaseshifts[instruction.target] = instruction.phase
        self.previous_phase_instruction_is_phase_set[instruction.target] = True

    @run.register(Pulse)
    @run.register(Acquire)
    def _(self, instruction: Pulse | Acquire):
        target = instruction.target
        emit = True
        if self.previous_phase_instruction_is_phase_set[target]:
            op_class = PhaseSet
        elif not np.isclose(self.accum_phaseshifts[target] % (2 * np.pi), 0.0):
            op_class = PhaseShift
        else:
            emit = False

        if emit:
            phase = self.accum_phaseshifts.pop(target)
            self.optimized_instructions.append(op_class(target=target, phase=phase))
            self.previous_phase_instruction_is_phase_set[target] = False

        self.optimized_instructions.append(instruction)

    @run.register(Delay)
    @run.register(Synchronize)
    def _(self, instruction: Delay | Synchronize):
        for target in instruction.targets:
            if self.previous_phase_instruction_is_phase_set[target]:
                self.optimized_instructions.append(
                    PhaseSet(target=target, phase=self.accum_phaseshifts.pop(target))
                )
                self.previous_phase_instruction_is_phase_set[target] = False
        self.optimized_instructions.append(instruction)


class PostProcessingSanitisation(TransformPass):
    """Checks that the :class:`PostProcessing` instructions that follow an acquisition are
    suitable for the acquisition mode, and removes them if not.
    """

    def run(
        self,
        ir: InstructionBuilder,
        _: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param met_mgr: The metrics manager to store the number of instructions after
            optimisation.
        """

        acquire_mode_output_var_map = {}
        discarded_pp = []
        for instr in ir:
            if isinstance(instr, Acquire):
                if instr.mode == AcquireMode.RAW:
                    raise ValueError(
                        "Invalid acquire mode. The target machine does not support "
                        "a RAW acquire mode."
                    )

                if instr.output_variable:
                    acquire_mode_output_var_map[instr.output_variable] = instr.mode

            elif isinstance(instr, PostProcessing):
                acq_mode = acquire_mode_output_var_map.get(instr.output_variable, None)

                if not acq_mode:
                    log.warning(
                        f"Post-processing output variable {instr.output_variable} is not associated with any acquire output variable."
                    )
                    discarded_pp.append(instr)

                else:
                    if not self._valid_pp(acq_mode, instr):
                        discarded_pp.append(instr)

        ir.instructions = [instr for instr in ir.instructions if instr not in discarded_pp]
        met_mgr.record_metric(
            MetricsType.OptimizedInstructionCount, ir.number_of_instructions
        )
        return ir

    def _valid_pp(self, acquire_mode: AcquireMode, pp: PostProcessing) -> bool:
        """
        Validate whether the post-processing instruction is valid with a given
        acquire mode.
        """

        if acquire_mode == AcquireMode.SCOPE:
            if (
                pp.process_type == PostProcessType.MEAN
                and ProcessAxis.SEQUENCE in pp.axes
                and len(pp.axes) <= 1
            ):
                return False

        elif acquire_mode == AcquireMode.INTEGRATOR:
            if (
                pp.process_type == PostProcessType.MEAN
                and ProcessAxis.TIME in pp.axes
                and len(pp.axes) <= 1
            ):
                return False
        return True


class MeasurePhaseResetSanitisation(TransformPass):
    """
    Adds a phase reset before every measure pulse.
    """

    def __init__(self, hardware_model: PhysicalHardwareModel):
        """
        :param hardware_model: The hardware model that holds calibrated information on the qubits on the QPU.
        """
        self.model = hardware_model
        self.measure_pulse_channels = self._get_measure_pulse_channels(hardware_model)

    @staticmethod
    def _get_measure_pulse_channels(hardware_model: PhysicalHardwareModel):
        """
        Returns a list of pulse channels that are used for measure pulses.
        """
        measure_pulse_channels = {}
        for qubit_id, qubit in hardware_model.qubits.items():
            measure_pulse_channels[qubit_id] = qubit.measure_pulse_channel.uuid
        return measure_pulse_channels

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        new_instructions = []
        for instr in ir.instructions:
            if isinstance(instr, MeasureBlock):
                for qubit_target in instr.qubit_targets:
                    new_instructions.append(
                        PhaseReset(target=self.measure_pulse_channels[qubit_target])
                    )
            elif (
                isinstance(instr, Pulse)
                and instr.target in self.measure_pulse_channels.values()
            ):
                new_instructions.append(PhaseReset(target=instr.target))
            new_instructions.append(instr)

        ir.instructions = new_instructions
        return ir


class InactivePulseChannelSanitisation(TransformPass):
    """Removes instructions that act on inactive pulse channels.

    Many channels that aren't actually needed to execute the program contain instructions,
    mainly :class:`Synchronize` instructions and :class:`PhaseShift` instructions which
    can happen when either of the instructions are applied to qubits. To simplify analysis
    and optimisations in later passes, it makes sense to filter these out to reduce the
    overall instruction amount.

    .. note::

        This pass requires results from the
        :class:`ActivePulseChannelAnalysis <qat.middleend.passes.analysis.ActivePulseChannelAnalysis>`
        to be stored in the results manager.
    """

    def run(
        self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        active_channels = set(res_mgr.lookup_by_type(ActivePulseChannelResults).targets)
        instructions: list[Instruction] = []
        for inst in ir.instructions:
            if isinstance(inst, Synchronize):
                # inactive channels need stripping from syncs
                targets = [target for target in inst.targets if target in active_channels]
                if len(targets) > 1:
                    inst.targets = targets
                    instructions.append(inst)
            elif isinstance(inst, QuantumInstruction):
                # other instructions need their targets checking
                if inst.target in active_channels:
                    instructions.append(inst)
            else:
                instructions.append(inst)
        ir.instructions = instructions

        # TODO: To be fixed with a module passed down the pipeline. (COMPILER-794)
        ir._pulse_channels = {
            k: v for k, v in ir._pulse_channels.items() if k in active_channels
        }
        return ir


class InstructionGranularitySanitisation(TransformPass):
    """Rounds the durations of quantum instructions so they are multiples of the clock
    cycle.

    Only supports quantum instructions with static non-zero durations. Assumes that
    instructions with a non-zero duration only act on a single pulse channel. The
    santisiation is done for all instructions simultaneously using numpy for performance.

    For :class:`Pulse` instructions with :class:`SampledWaveform` waveforms, the durations
    are rounded up by padding the pulse with zero amplitude at the end. For other relevant
    instructions, we round down: this is for compatibility with calibration files that are
    calibrated using legacy code. However, in the future we might consider changing this
    to round up for consistency.

    .. warning::

        This pass has the potential to invalidate the timings for sequences of instructions
        that are time-sensitive. For example, if a pulse has an invalid time, it will round
        it up to the nearest integer multiple. Furthemore, it will assume that
        :class:`Acquire` instructions have no delay. This can be forced explicitly using the
        :class:`AcquireSanitisation` pass.
    """

    def __init__(self, target_data: TargetData):
        """:param target_data: Target-related information."""

        self.clock_cycle = target_data.clock_cycle

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        quantum_instructions = []
        custom_pulses = []
        for inst in ir.instructions:
            if isinstance(inst, (Acquire, Delay)) or (
                isinstance(inst, Pulse) and not isinstance(inst.waveform, SampledWaveform)
            ):
                quantum_instructions.append(inst)
            elif isinstance(inst, Pulse) and isinstance(inst.waveform, SampledWaveform):
                custom_pulses.append(inst)

        self._sanitise_quantum_instructions(quantum_instructions)
        self._sanitise_custom_pulses(custom_pulses)

        return ir

    def _clock_cycle_multiples(
        self, instructions: list[Pulse | Acquire | Delay]
    ) -> NDArray[float]:
        """
        Extracts the number of clock cycles for each instruction
        """
        durations = np.asarray([inst.duration for inst in instructions])
        return durations / self.clock_cycle

    def _sanitise_quantum_instructions(self, instructions: list[Pulse | Acquire | Delay]):
        """Sanitises the durations quantum instructions with non-zero duration by rounding
        down to the nearest clock cycle."""

        multiples = self._clock_cycle_multiples(instructions)
        # 1e-10 for floating point errors
        rounded_multiples = np.floor(multiples + 1e-10).astype(int)
        durations_equal = np.isclose(multiples, rounded_multiples)

        invalid_instructions: set[str] = set()
        for idx in np.where(np.logical_not(durations_equal))[0]:
            inst = instructions[idx]
            invalid_instructions.add(str(inst))
            new_duration = rounded_multiples[idx] * self.clock_cycle
            if isinstance(inst, Pulse):
                inst.update_duration(new_duration)
            else:
                if isinstance(inst, Acquire) and inst.filter is not None:
                    self._sanitise_acquire_pulse(inst, new_duration)
                inst.duration = new_duration

        if len(invalid_instructions) >= 1:
            log.info(
                "The following instructions do not have durations that are integer "
                f"multiples of the clock cycle {self.clock_cycle}, and will be rounded "
                "down: " + ", ".join(set(invalid_instructions))
            )

    @staticmethod
    def _sanitise_acquire_pulse(instruction: Acquire, new_duration: float):
        """
        Acquire instructions with filters have a duration equal to the filter duration,
        so we need to update the filter as well.
        """
        if (
            isinstance(instruction.filter.waveform, SampledWaveform)
            and new_duration < instruction.duration
        ):
            # This is a temporary workaround due to the fact that the acquire duration gets rounded
            # down to the nearest clock cycle, which results in cutting off the sample.
            # TODO: Review for COMPILER-488 changes.
            n_samples = int(
                np.floor(new_duration / instruction.filter.waveform.sample_time)
            )
            instruction.filter.waveform.samples = instruction.filter.waveform.samples[
                :n_samples
            ]
            instruction.filter.duration = new_duration
        instruction.filter.update_duration(new_duration)

    def _sanitise_custom_pulses(self, instructions: list[Pulse]):
        """Sanitises the durations of :class:`SampledWaveform`s by padding the
        waveforms with zero amplitudes."""

        multiples = self._clock_cycle_multiples(instructions)
        # 1e-10 for floating point errors
        rounded_multiples = np.ceil(multiples - 1e-10).astype(int)
        durations_equal = np.isclose(multiples, rounded_multiples)

        invalid_instructions: set[str] = set()
        for idx in np.where(np.logical_not(durations_equal))[0]:
            inst = instructions[idx]
            invalid_instructions.add(str(inst))
            new_duration = rounded_multiples[idx] * self.clock_cycle
            inst.update_duration(new_duration)

        if len(invalid_instructions) > 1:
            log.info(
                "The following sampled waveform pulses do not have durations that are integer "
                f"multiples of the clock cycle {self.clock_cycle}, and will be rounded "
                "up by padding with zero amplitudes: "
                + ", ".join(set(invalid_instructions))
            )


class InstructionLengthSanitisation(TransformPass):
    """
    Checks if quantum instructions are too long and splits if necessary.
    """

    def __init__(self, model: PhysicalHardwareModel, target_data: TargetData):
        """
        :param duration_limit: The maximum allowed clock cycles per instruction.

        .. warning::

            The pass will assume that the durations of instructions are sanitised to the
            granularity of the pulse channels. If instructions that do not meet the criteria are
            provided, it might produce incorrect instructions (i.e., instructions that are shorter than
            the clock cycle). This can be enforced using the :class:`InstructionGranularitySanitisation <qat.middleend.passes.transform.InstructionGranularitySanitisation>`
            pass.
        """
        self.duration_limit = target_data.QUBIT_DATA.pulse_duration_max
        self.sample_times = self._create_sample_time_map(model, target_data)

    @staticmethod
    def _create_sample_time_map(
        model: PhysicalHardwareModel, target_data: TargetData
    ) -> dict[str, float]:
        """Maps physical channels onto their sample times."""
        channel_data = {}
        for qubit in model.qubits.values():
            channel_data[qubit.physical_channel.uuid] = target_data.QUBIT_DATA.sample_time
            channel_data[qubit.resonator.physical_channel.uuid] = (
                target_data.RESONATOR_DATA.sample_time
            )
        return channel_data

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        new_instructions = []
        for instr in ir.instructions:
            if isinstance(instr, Delay) and instr.duration > self.duration_limit:
                new_instructions.extend(self._batch_delay(instr, self.duration_limit))

            elif (
                isinstance(instr, Pulse)
                and isinstance(instr.waveform, SquareWaveform)
                and instr.waveform.width > self.duration_limit
            ):
                new_instructions.extend(
                    self._batch_square_pulse(instr, self.duration_limit)
                )

            elif (
                isinstance(instr, Pulse)
                and isinstance(instr.waveform, SampledWaveform)
                and instr.duration > self.duration_limit
            ):
                target = next(iter(instr.targets))
                physical_channel_id = ir.get_pulse_channel(target).physical_channel_id
                sample_time = self.sample_times[physical_channel_id]
                new_instructions.extend(
                    self._batch_custom_waveform(instr, self.duration_limit, sample_time)
                )

            else:
                new_instructions.append(instr)
        ir.instructions = new_instructions
        return ir

    @staticmethod
    def _batch_delay(instruction: Delay, max_duration: float):
        n_instr = int(instruction.duration // max_duration)
        remainder = instruction.duration % max_duration

        batch_instr = []
        for _ in range(n_instr):
            batch_instr.append(Delay(target=instruction.targets, duration=max_duration))

        if remainder > 0.0:
            batch_instr.append(Delay(target=instruction.targets, duration=remainder))

        return batch_instr

    @staticmethod
    def _batch_square_pulse(instruction: Pulse, max_width: float):
        n_instr = int(instruction.waveform.width // max_width)
        remainder = instruction.waveform.width % max_width

        batch_instr = []
        pulse = deepcopy(instruction)
        pulse.update_duration(max_width)
        for _ in range(n_instr):
            batch_instr.append(deepcopy(pulse))

        if remainder > 0.0:
            pulse = deepcopy(pulse)
            pulse.update_duration(remainder)
            batch_instr.append(pulse)

        return batch_instr

    def _batch_custom_waveform(
        self, instruction: Pulse, max_duration: float, sample_time: float
    ):
        """Breaks up a custom pulse into multiple custom pulses with none exceeding the
        maximum duration."""

        n_instr = int(instruction.duration // max_duration)
        remainder = instruction.duration % max_duration
        max_samples = int(round(max_duration / sample_time))

        batch_instr = []
        for i in range(n_instr):
            waveform = SampledWaveform(
                samples=instruction.waveform.samples[
                    i * max_samples : (i + 1) * max_samples
                ],
                sample_time=instruction.waveform.sample_time,
            )
            pulse = Pulse(
                targets=instruction.target,
                waveform=waveform,
                ignore_channel_scale=instruction.ignore_channel_scale,
                duration=max_duration,
            )
            batch_instr.append(pulse)

        if remainder:
            num_samples = int(round(remainder / sample_time))
            waveform = SampledWaveform(
                samples=instruction.waveform.samples[-num_samples:],
                sample_time=instruction.waveform.sample_time,
            )
            pulse = Pulse(
                targets=instruction.target,
                waveform=waveform,
                ignore_channel_scale=instruction.ignore_channel_scale,
                duration=remainder,
            )
            batch_instr.append(pulse)

        return batch_instr


class ReturnSanitisation(TransformPass):
    """Squashes all :class:`Return` instructions into a single one. Adds a :class:`Return`
    with all acquisitions if none is found."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        returns = [inst for inst in ir if isinstance(inst, Return)]
        measure_blocks = [inst for inst in ir if isinstance(inst, MeasureBlock)]

        if returns:
            unique_variables = set(itertools.chain(*[ret.variables for ret in returns]))
            for ret in returns:
                ir.instructions.remove(ret)
        else:
            # If we do not have an explicit return, imply all results.
            unique_variables = set(
                itertools.chain(*[mb.output_variables for mb in measure_blocks])
            )

        ir.returns(list(unique_variables))
        return ir


class RepeatSanitisation(TransformPass):
    """Adds repeat counts and repetition periods to :class:`Repeat` instructions. If none
    is found, a repeat instruction is added."""

    def __init__(self, target_data: TargetData):
        """
        :param target_data: Target-related information.
        """
        self.default_shots = target_data.default_shots

    def run(
        self,
        ir: InstructionBuilder,
        *args,
        compiler_config: CompilerConfig = None,
        **kwargs,
    ):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        num_shots = self.default_shots
        configured = False
        if compiler_config is not None and compiler_config.repeats is not None:
            configured = True
            num_shots = compiler_config.repeats

        for inst in ir.instructions:
            if isinstance(inst, Repeat):
                if inst.repeat_count is None:
                    inst.repeat_count = num_shots
                elif configured and inst.repeat_count != num_shots:
                    raise ValueError(
                        f"Repeat instruction 'repeat_count' [{inst.repeat_count}] not "
                        f"matching CompilerConfig value [{num_shots}]."
                    )
                return ir

        ir.repeat(num_shots)
        log.warning(
            "Could not find any repeat instructions. "
            f"One has been added with {num_shots} shots."
        )
        return ir


class BatchedShots(TransformPass):
    """Determines how shots should be grouped when the total number exceeds that maximum
    allowed.

    The target machine might have an allowed number of shots that can be executed by a
    single execution call. To execute a number of shots greater than this value, shots can
    be batched, with each batch executed by its own "execute" call on the target machine.
    For example, if the maximum number of shots for a target machine is 2000, but you
    required 4000 shots, then this could be done as [2000, 2000] shots.

    Now consider the more complex scenario where  4001 shots are required. Clearly this can
    be done in three batches. While it is tempting to do this in batches of [2000, 2000, 1],
    for some target machines, specification of the number of shots can only be achieved at
    compilation (as opposed to runtime). Batching as described above would result in us
    needing to compile two separate programs. Instead, it makes more sense to batch the
    shots as three lots of 1334 shots, which gives a total of 4002 shots. The extra two
    shots can just be discarded at run time.

    .. warning::

        This pass makes certain assumptions about the IR, including that there is only
        at most one :class:`Repeat` instruction that contributes to the number of readouts,
        and there is at most one readout per shot. It will change the number of shots in
        the :class:`Repeat` instruction to the decided batch size, and store the total
        required shots in the IR, which is less than ideal. It would be nice to save this
        as an analysis result, but an unfortante side effect of the decoupled nature of
        the middle- and back-end is that the results manager might not necessarily be
        passed along.
    """

    def __init__(self, target_data: TargetData):
        """
        :param target_data: Target-related information.
        """
        self.max_shots = target_data.max_shots

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        repeats = [inst for inst in ir if isinstance(inst, Repeat)]
        if len(repeats) == 0:
            return ir

        if len(repeats) > 1:
            log.warning(
                "Multiple repeat instructions found. Using only the first for batching of "
                "shots."
            )

        repeat = repeats[0]
        num_shots = repeat.repeat_count
        num_batches = int(np.ceil(num_shots / self.max_shots))
        shots_per_batch = int(np.ceil(num_shots / num_batches))

        if num_batches > 1:
            log.info(
                f"The number of shots {num_shots} exceeds the maximum allowed on the "
                f"target. Batching as {num_batches} batches of {shots_per_batch} shots."
            )

        repeat.repeat_count = shots_per_batch
        ir.shots = num_shots
        ir.compiled_shots = shots_per_batch
        return ir


class ResetsToDelays(TransformPass):
    """
    Transforms :class:`Reset` operations to :class:`Delay`s.

    Note that the delays do not necessarily agree with the granularity of the underlying target machine.
    This can be enforced using the :class:`InstructionGranularitySanitisation` pass.
    """

    def __init__(self, model: PhysicalHardwareModel, target_data: TargetData):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
        :param target_data: Target-related information.
        """
        self.model = model
        self.passive_reset_time = target_data.QUBIT_DATA.passive_reset_time

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        *args,
        compiler_config: CompilerConfig = None,
        **kwargs,
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        :param compiler_config: The compiler configuration.
        """
        active_pulse_channels = res_mgr.lookup_by_type(ActivePulseChannelResults)

        reset_time = self._get_reset_time(compiler_config)

        new_instructions = []
        for instr in ir:
            if isinstance(instr, Reset):
                qubit = self.model.qubits[instr.qubit_target]
                for pulse_ch in active_pulse_channels.from_qubit(qubit):
                    new_instructions.append(
                        Delay(
                            target=pulse_ch.uuid,
                            duration=reset_time,
                        )
                    )

            else:
                new_instructions.append(instr)

        ir.instructions = new_instructions
        return ir

    def _get_reset_time(
        self,
        compiler_config: CompilerConfig,
    ) -> float:
        """Gets the reset time for a qubit from the compiler configuration, or falls back to
        the passive reset time from the target data."""

        if getattr(compiler_config, "passive_reset_time", None) is not None:
            return compiler_config.passive_reset_time

        if getattr(compiler_config, "repetition_period", None) is not None:
            log.warning(
                "The `repetition_period` in `CompilerConfig` is deprecated. "
                "Please use `passive_reset_time` instead. "
                f"Using the default `passive_reset_time` {self.passive_reset_time}."
            )

        return self.passive_reset_time


class SquashDelaysOptimisation(TransformPass):
    """Looks for consecutive :class:`Delay` instructions on a pulse channel and squashes
    them into a single instruction.

    Because :class:`Synchronize` instructions across multiple pulse channels are used so
    frequently to ensure pulses play at the correct timings, it means we can have sequences
    of many delays. Reducing the number of delays will simplify timing analysis later in
    the compilation.

    :class:`Delay` instructions commute with phase related instructions, so the only
    instructions that separate delays in a meaningful way are: :class:`Pulse`:,
    :class:`CustomPulse` and :class:`Acquire` instructions. We also need to be careful to
    not squash delays that contain a variable time.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        :param met_mgr: The metrics manager to store the number of instructions after
            optimisation.
        """
        delimiter_types = (
            Acquire,
            Assign,
            Delay,
            Jump,
            Label,
            PhaseReset,
            PhaseSet,
            Pulse,
        )
        accumulated_delays: dict[str, float] = defaultdict(float)
        instructions: list[Instruction] = []
        for inst in ir:
            if isinstance(inst, Delay):
                accumulated_delays[inst.target] += inst.duration

            elif isinstance(inst, delimiter_types):
                if isinstance(inst, QuantumInstruction):
                    targets = {inst.target}
                else:
                    targets = accumulated_delays.keys()
                for target in targets:
                    if (duration := accumulated_delays[target]) > 0.0:
                        instructions.append(Delay(target=target, duration=duration))
                        accumulated_delays[target] = 0.0
                instructions.append(inst)
            else:
                instructions.append(inst)

        for target, duration in accumulated_delays.items():
            if duration != 0.0:
                instructions.append(Delay(target=target, duration=duration))
        ir.instructions = instructions

        met_mgr.record_metric(MetricsType.OptimizedInstructionCount, len(instructions))
        return ir


class ScopeSanitisation(TransformPass):
    """Bubbles up all sweeps and repeats to the beginning of the list and adds delimiter
    instructions to the repeats and sweeps signifying the end of their scopes.

    .. warning::

        This pass is intended for use with legacy builders that do not use
        :class:`EndRepeat` and :class:`EndSweep` instructions. It will add a delimiter
        instruction for each :class:`Repeat` and :class:`Sweep` found in the IR.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        tail, head = partition(lambda inst: isinstance(inst, Repeat), ir.instructions)
        tail, head = list(tail), list(head)

        ir.instructions = head + tail
        return ir


class EndOfTaskResetSanitisation(TransformPass):
    """Checks for a reset on each active qubit at the end of a task, and adds Reset
    operations if not found.

    After each shot, it is expected that the qubit is returned to its ground state, ready
    for the next shot. This pass ensures this is the case by checking if the last "active"
    operation on an qubit is a :class:`Reset`, and if not, adds a :class:`Reset` to the end
    of the instruction list.
    """

    def __init__(self, model: PhysicalHardwareModel):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
        """
        self.model = model

    def run(
        self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        active_pulse_channels = res_mgr.lookup_by_type(ActivePulseChannelResults)
        qubit_map: dict[Qubit, None | bool] = {
            qubit: None for qubit in active_pulse_channels.qubits
        }

        for inst in reversed(ir.instructions):
            if not isinstance(inst, (Pulse, Acquire, Reset)):
                continue

            elif isinstance(inst, Reset):
                # If a qubit only sees a reset, its not "active", so ignore.
                qubit = self.model.qubits[inst.qubit_target]
                if qubit in qubit_map and qubit_map[qubit] is None:
                    qubit_map[qubit] = True
            else:
                qubit = active_pulse_channels.pulse_channel_to_qubit_map[inst.target]
                if qubit_map[qubit] is None:
                    qubit_map[qubit] = False

            if all([valid_reset is not None for valid_reset in qubit_map.values()]):
                break

        for qubit, valid_reset in qubit_map.items():
            if not valid_reset:
                ir.add(Reset(qubit_target=self.model.index_of_qubit(qubit)))

        return ir


class FreqShiftSanitisation(TransformPass):
    """
    Looks for any active frequency shift pulse channels in the hardware model and adds
    square pulses for the duration.

    .. warning::

        This pass assumes the following, which is achieved via other passes:

        * :class:`Synchronize` instructions have already been lowered to :class:`Delay`
          instructions.
        * Durations are static.
    """

    def __init__(self, model: PhysicalHardwareModel):
        """:param model: The hardware model containing the frequency shift channels."""

        self.model = model
        self.active_freq_shift_pulse_channels = self.get_active_freq_shift_pulse_channels(
            model
        )

    def run(
        self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> InstructionBuilder:
        """
        :param ir: The QatIR as an instruction builder.
        :param res_mgr: The results manager.
        """

        freq_shift_channels = self.active_freq_shift_pulse_channels
        if len(freq_shift_channels) == 0:
            return ir
        ir = self.add_freq_shift_to_ir(ir, set(freq_shift_channels.keys()))

        if res_mgr.check_for_type(ActivePulseChannelResults):
            res = res_mgr.lookup_by_type(ActivePulseChannelResults)
            for pulse_channel, qubit in freq_shift_channels.items():
                res.add_target(pulse_channel, qubit)

        return ir

    @staticmethod
    def get_active_freq_shift_pulse_channels(model) -> dict[FreqShiftPulseChannel, Qubit]:
        """Returns all active frequency shift pulse channels found in the hardware model."""
        pulse_channels = dict()
        for qubit in model.qubits.values():
            freq_shift_pulse_ch = qubit.freq_shift_pulse_channel
            if freq_shift_pulse_ch.active:
                pulse_channels[freq_shift_pulse_ch] = qubit

        return pulse_channels

    @staticmethod
    def add_freq_shift_to_ir(
        ir: InstructionBuilder, freq_shift_channels: set[PulseChannel]
    ) -> InstructionBuilder:
        """Adds frequency shift instructions to the instruction builder."""

        durations = defaultdict(float)
        idx = None  # Index of first quantum instruction.
        new_instructions = []
        for i, inst in enumerate(ir):
            if isinstance(inst, QuantumInstruction):
                if idx is None:
                    idx = i

                if (duration := inst.duration) > 0:
                    durations[inst.target] += duration

            new_instructions.append(inst)

        # If no quantum instructions found, return the IR as is.
        if len(durations) == 0:
            return ir

        max_duration = max(durations.values())
        for pulse_ch in freq_shift_channels:
            pulse = Pulse(
                target=pulse_ch.uuid,
                waveform=SquareWaveform(
                    amp=pulse_ch.amp, width=max_duration, phase=pulse_ch.phase
                ),
            )
            new_instructions.insert(idx, pulse)

        ir.instructions = new_instructions
        return ir


class InitialPhaseResetSanitisation(TransformPass):
    """
    Checks if every active pulse channel has a phase reset in the beginning.

    .. warning::

        This pass implies that an `ActivePulseChannelAnalysis` is performed prior to this pass.
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        active_targets = res_mgr.lookup_by_type(ActivePulseChannelResults).targets

        if active_targets:
            index = next(
                filter(
                    lambda tup: True if isinstance(tup[1], QuantumInstruction) else None,
                    enumerate(ir.instructions),
                )
            )[0]
            head, tail = ir.instructions[:index], ir.instructions[index:]
            resets = [PhaseReset(target=target) for target in active_targets]
            ir.instructions = head + resets + tail

        return ir


class LowerSyncsToDelays(TransformPass):
    """Lowers :class:`Synchronize` instructions to :class:`Delay` instructions with static
    times.

    Increments through the instruction list, keeping track of the cumulative duration.
    When :class:`Synchronize` instructions are encountered, it is replaced with
    :class:`Delay` instructions with timings calculated from the cumulative durations.

    .. warning::

        Any manipulations of the instruction set that will alter the timeline and occur
        after this pass could invalidate the intention of the :class:`Synchronize`
        instruction.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        durations: dict[str, float] = defaultdict(float)
        new_instructions: list[Instruction] = []

        for inst in ir.instructions:
            self.process_instruction(inst, new_instructions, durations)

        ir.instructions = new_instructions
        return ir

    @singledispatchmethod
    def process_instruction(
        self, instruction, new_instructions: list[Instruction], durations: dict[str, float]
    ):
        raise NotImplementedError(
            f"No processing method for instruction type {type(instruction)}"
        )

    @process_instruction.register(Instruction)
    def _(
        self,
        instruction: Instruction,
        new_instructions: list[Instruction],
        durations: dict[str, float],
    ):
        """Default handler for instructions that do not have a specific processing
        method."""
        new_instructions.append(instruction)

    @process_instruction.register(QuantumInstruction)
    def _(
        self,
        instruction: QuantumInstruction,
        new_instructions: list[Instruction],
        durations: dict[str, float],
    ):
        """Processes QuantumInstructions by updating their target durations."""
        durations[instruction.target] += instruction.duration
        new_instructions.append(instruction)

    @process_instruction.register(Synchronize)
    def _(
        self,
        instruction: Synchronize,
        new_instructions: list[Instruction],
        durations: dict[str, float],
    ):
        """Process Synchronize instructions by converting them to Delay instructions."""
        targets = instruction.targets
        current_durations = np.asarray([durations[target] for target in targets])
        max_duration = np.max(current_durations)
        sync_durations = max_duration - current_durations
        delay_instrs = [
            Delay(target=target, duration=sync_durations[i])
            for i, target in enumerate(targets)
            if sync_durations[i] > 0.0
        ]
        new_instructions.extend(delay_instrs)
        durations.update({target: max_duration for target in targets})


class EvaluateWaveforms(TransformPass):
    """Evaluates the amplitudes of :class:`Waveform`s within :class:`Pulse` instructions,
    replacing them with a :class:`SampledWaveform` and accounting for the scale of the
    pulse channel.

    :class:`Waveform` dataclasses are defined by (often many) parameters. With the exception
    of specific shapes, they cannot be implemented directly on hardware. Instead, we
    evaluate the waveform at discrete times, and communicate these values to the target.
    This pass evaluates pulses early in the compilation pipeline.

    The resulting :class:`Pulse` instructions will have :code:`ignore_channel_scale=True`.
    """

    def __init__(
        self,
        model: PhysicalHardwareModel,
        target_data: TargetData,
        ignored_shapes: tuple[type] | None = None,
        acquire_ignored_shapes: tuple[type] | None = None,
    ):
        """
        :param model: The hardware model that holds calibrated information on the qubits,
            which is required to extract the scale factor of pulse channels.
        :param target_data: Target-related information, which is used to extract the sample
            time of the pulse channels.
        :param ignored_shapes: A list of pulse shapes that are not evaluated to a custom
            pulse, defaults to `[PulseShapeType.SQUARE]`.
        :param acquire_ignored_shapes: A list of pulse shapes that are not evaluated to a
            custom pulse for :class:`Acquire` filters, defaults to `[]`.
        """

        self.ignored_shapes = self._sanitise_ignored_shapes(ignored_shapes, SquareWaveform)
        self.acquire_ignored_shapes = self._sanitise_ignored_shapes(
            acquire_ignored_shapes, ()
        )
        self.sample_times = self._extract_pulse_channel_features(model, target_data)

    @staticmethod
    def _sanitise_ignored_shapes(
        ignored_shapes: tuple[type] | type, default: tuple[type] | type
    ) -> tuple[type]:
        """Sanitises the ignored shapes to ensure they are a list of types."""
        ignored_shapes = ignored_shapes if ignored_shapes is not None else default
        return ignored_shapes if isinstance(ignored_shapes, tuple) else (ignored_shapes,)

    @staticmethod
    def _extract_pulse_channel_features(
        model: PhysicalHardwareModel, target_data: TargetData
    ) -> dict[str, float]:
        """Extracts the  sample rate for all pulse channels as dicts, with physical channel
        IDs as keys."""

        sample_times = {}
        for qubit in model.qubits.values():
            sample_times[qubit.physical_channel.uuid] = target_data.QUBIT_DATA.sample_time
            sample_times[qubit.resonator.physical_channel.uuid] = (
                target_data.RESONATOR_DATA.sample_time
            )

        return sample_times

    def run(
        self, ir: QuantumInstructionBuilder, *args, **kwargs
    ) -> QuantumInstructionBuilder:
        """:param ir: The list of instructions as an instruction builder."""

        waveform_lookup = defaultdict(dict)
        instructions = []
        for inst in ir.instructions:
            new_inst = self.process_instruction(inst, ir, waveform_lookup=waveform_lookup)
            instructions.append(new_inst)
        ir.instructions = instructions
        return ir

    @singledispatchmethod
    def process_instruction(
        self, instruction: Instruction, ir: QuantumInstructionBuilder, **kwargs
    ):
        """Default handler for instructions that do not have a waveform evaluation."""
        return instruction

    @process_instruction.register(Acquire)
    def _(self, instruction: Acquire, ir: QuantumInstructionBuilder, **kwargs) -> Acquire:
        """Implements the waveform evaluation on the filter for an :class:`Acquire`
        instruction."""

        instruction.filter = self.process_instruction(
            instruction.filter, ir, ignored_shapes=self.acquire_ignored_shapes, **kwargs
        )
        return instruction

    @process_instruction.register(Pulse)
    def _(self, instruction: Pulse, ir: QuantumInstructionBuilder, **kwargs) -> Pulse:
        """Evaluates the waveform within the :class:`Pulse` and handles the channel scale
        factor if required."""

        pulse_channel = ir.get_pulse_channel(instruction.target)
        scale = pulse_channel.scale if not instruction.ignore_channel_scale else 1.0

        instruction.waveform = self.evaluate_waveform(
            instruction.waveform,
            ignored_shapes=kwargs.pop("ignored_shapes", self.ignored_shapes),
            target=instruction.target,
            physical_channel=pulse_channel.physical_channel_id,
            scale=scale,
            **kwargs,
        )
        instruction.ignore_channel_scale = True
        return instruction

    @singledispatchmethod
    def evaluate_waveform(self, waveform, *args, **kwargs):
        """Default handler for waveform evaluation that does not have a specific
        implementation."""
        return ValueError(
            f"No waveform evaluation method for waveform type {type(waveform)}."
        )

    @evaluate_waveform.register(Waveform)
    def _(
        self,
        waveform: Waveform,
        ignored_shapes: list[type],
        waveform_lookup: dict[str, dict[str, NDArray]],
        target: str,
        physical_channel: str,
        scale: complex | float = 1.0 + 0.0j,
    ) -> Waveform | SampledWaveform:
        """Implements the waveform evaluation for :class:`Waveform`s."""

        # We cannot bake the waveform type into singledispatch here as the ignored shapes
        # is dynamic..
        if isinstance(waveform, ignored_shapes):
            if np.isclose(scale, 1.0):
                return waveform
            else:
                # We reinstantiate a new waveform as a preacaution against accidently
                # accidently shared instances.
                attrs = waveform.model_dump()
                attrs["amp"] *= scale
                return type(waveform)(**attrs)

        samples = waveform_lookup[target].get(waveform, None)
        if samples is not None:
            new_waveform = SampledWaveform(samples=samples)
        else:
            # Evaluate the waveform at discrete times
            sample_time = self.sample_times[physical_channel]
            edge = waveform.duration / 2.0 - sample_time * 0.5
            num_samples = int(np.ceil(waveform.duration / sample_time - 1e-10))
            t = np.linspace(start=-edge, stop=edge, num=num_samples)
            new_waveform = waveform.sample(t)
            waveform_lookup[target][waveform] = new_waveform.samples

        if not np.isclose(scale, 1.0):
            new_waveform.samples = new_waveform.samples * scale
        return new_waveform

    @evaluate_waveform.register(SampledWaveform)
    def _(
        self, waveform: SampledWaveform, scale: complex | float = 1.0 + 0.0j, **kwargs
    ) -> SampledWaveform:
        """Multiplies the waveform by a scale factor if required."""

        if not np.isclose(scale, 1.0):
            waveform.samples = waveform.samples * scale
        return waveform


class SynchronizeTask(TransformPass):
    """Synchronizes all active pulse channels in a task.

    Adds a synchronize to the end of the instruction list for all active pulse channels,
    which is extracted from the :class:`ActivePulseChannelAnalysis` pass. This is useful to
    do before resetting the qubits, as it ensures no qubit is reset while the task is still
    "active", reducing the effects of cross-talk.
    """

    def run(
        self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        active_channels = res_mgr.lookup_by_type(ActivePulseChannelResults).targets
        if len(active_channels) > 1:
            ir.add(Synchronize(targets=active_channels))
        return ir


class AcquireSanitisation(TransformPass):
    """Sanitises the :class:`Acquire` instruction to separate the :class:`Delay` from
    :class:`Acquire` instructions.

    :class:`Acquire` instructions are defined by a "duration" for which they instruct the
    target to readout. They also contain a "delay" attribute, which instructions the
    acquisition to start after some given time. This pass separates acqusitions with a
    delay into two instructions for the first acquire that acts on the channel.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        new_instructions: list[Instruction] = []

        for inst in ir:
            if isinstance(inst, Acquire):
                if inst.delay:
                    delay = Delay(target=inst.target, duration=inst.delay)
                    inst.delay = 0.0
                    new_instructions.extend([delay, inst])
                else:
                    new_instructions.append(inst)
            else:
                new_instructions.append(inst)
        ir.instructions = new_instructions
        return ir


PydPhaseOptimisation = PhaseOptimisation
PydPostProcessingSanitisation = PostProcessingSanitisation
PydReturnSanitisation = ReturnSanitisation
PydRepeatSanitisation = RepeatSanitisation
PydBatchedShots = BatchedShots
PydResetsToDelays = ResetsToDelays
PydSquashDelaysOptimisation = SquashDelaysOptimisation
PydRepeatTranslation = RepeatTranslation
PydMeasurePhaseResetSanitisation = MeasurePhaseResetSanitisation
PydInactivePulseChannelSanitisation = InactivePulseChannelSanitisation
PydInstructionGranularitySanitisation = InstructionGranularitySanitisation
PydInstructionLengthSanitisation = InstructionLengthSanitisation
PydScopeSanitisation = ScopeSanitisation
PydEndOfTaskResetSanitisation = EndOfTaskResetSanitisation
PydFreqShiftSanitisation = FreqShiftSanitisation
PydInitialPhaseResetSanitisation = InitialPhaseResetSanitisation
PydLowerSyncsToDelays = LowerSyncsToDelays
PydEvaluateWaveforms = EvaluateWaveforms
PydSynchronizeTask = SynchronizeTask
PydAcquireSanitisation = AcquireSanitisation



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from random import Random

from qat.model.convert_purr import convert_purr_echo_hw_to_pydantic
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.loaders.base import BasePhysicalModelLoader
from qat.purr.backends.echo import Connectivity
from qat.purr.compiler.hardware_models import ErrorMitigation
from qat.utils.hardware_model import (
    apply_setup_to_echo_hardware,
    ensure_connected_connectivity,
    random_connectivity,
    random_quality_map,
)

from .purr.echo import EchoModelLoader as LegacyEchoModelLoader


class EchoModelLoader(BasePhysicalModelLoader):
    def __init__(
        self,
        qubit_count: int = 4,
        connectivity: Connectivity | list[(int, int)] | None = None,
        error_mitigation: ErrorMitigation | None = None,
    ):
        self._legacy = LegacyEchoModelLoader(
            qubit_count=qubit_count, connectivity=connectivity
        )
        self._error_mitigation = error_mitigation

    def load(self) -> PhysicalHardwareModel:
        legacy_model = self._legacy.load()
        legacy_model.error_mitigation = self._error_mitigation
        return convert_purr_echo_hw_to_pydantic(legacy_model)


class JaggedEchoModelLoader(BasePhysicalModelLoader):
    def __init__(
        self,
        qubit_count: int = 4,
        qubit_indices: list[int] = None,
        connectivity: list[(int, int)] | None = None,
        error_mitigation: ErrorMitigation | None = None,
        random_seed: int | None = None,
    ):
        """Load a converted Echo hardware model with non-sequential physical qubit indices.

        :param qubit_count: Number of physical qubits to select.
        :param qubit_indices: Specific indices to select from, if used in combination
            with `connectivity` these indices must all be in the connectivity keys.
            Defaults to a selection of `qubit_count` indices in `connectivity` if
            provided else selected from a generated :func:`random_connectivity`.
        :param connectivity: List of tuples representing the qubit connections. Defaults
            to a genetared :func:`random_connectivity` including `qubit_indices` if
            provided.
        :param error_mitigation: Error mitigation settings to apply to the model.
        :param random_seed: Seed for random connectivity and quality map generation.
        """
        if isinstance(connectivity, Connectivity):
            raise NotImplementedError(
                "JaggedEchoModelLoader does not support Connectivity type."
            )

        random = Random(random_seed)
        if connectivity is None:
            if qubit_indices is None:
                n = qubit_count + 5
            else:
                n = max(qubit_indices) + 1
            connectivity = random_connectivity(n, seed=random)
        else:
            connectivity_dict = defaultdict(set)
            for qubit, coupled in connectivity:
                connectivity_dict[qubit].add(coupled)
            connectivity = connectivity_dict

        available_indices = set(connectivity)

        if qubit_indices is None:
            qubit_indices = random.sample(list(available_indices), qubit_count)
        else:
            if not (qi_set := set(qubit_indices)) <= available_indices:
                missing = qi_set - available_indices
                raise ValueError(
                    f"Incompatible connectivity and qubit_indices, chosen indices {missing}"
                    " are not in connectivity list."
                )

        new_connectivity = ensure_connected_connectivity(connectivity, qubit_indices)
        self._quality_map = random_quality_map(
            new_connectivity, min_quality=1, max_quality=100, seed=random
        )
        self._qubit_indices = set(qubit_indices)
        self._qubit_count = len(qubit_indices)
        self._error_mitigation = error_mitigation

    def load(self) -> PhysicalHardwareModel:
        legacy_model = apply_setup_to_echo_hardware(
            self._qubit_count,
            connectivity=self._quality_map,
            qubit_indices=self._qubit_indices,
        )
        legacy_model.error_mitigation = self._error_mitigation
        return convert_purr_echo_hw_to_pydantic(legacy_model)


PydEchoModelLoader = EchoModelLoader
